#![no_main]
#![no_std]
#[allow(unused_extern_crates)]
extern crate panic_halt;

use core::fmt::Write;
use rtfm::app;
use embedded_hal::adc::OneShot;

use lpc43xx_hal as hal;
use hal::{
    io::pin::{split_ports, PinSplit, IntoGpioPin},
    pac,
    serial::Serial,
    adc::{Adc, AdcIn, IntoAdc},
    timer::{Timer, InterruptType, EMC0_A, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
    clock::uart_pll,
};
use lpc43xx_hal::io::pin::{Pin, U4, U3};
use pac::usart0::lcr::{BC_A, DLAB_A, PE_A, PS_A, SBS_A, WLS_A};

const TIMER_PERIOD: u32 = 8192;

type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut ADC: Adc<pac::ADC0> = ();
    static mut TIMER: Timer0 = ();
    static mut USART: Serial<pac::USART3> = ();
    static mut ADC0_IN: Pin<U4, U3, AdcIn<hal::adc::ADC0>> = ();

    #[init]
    fn init() -> init::LateResources {
        cortex_m::asm::bkpt();
        
        // ========== PORT INITIALIZATION ==========
        let port = split_ports(device.SCU, device.GPIO_PORT);

        // USART3 I/O
        let port_9 = port.port_9.split_pins();
        let uart_tx = port_9.p9_3;
        let uart_rx = port_9.p9_4;

        // ADC0 chan 0 input
        let port_4 = port.port_4.split_pins();
        let adc0_in = unsafe { port_4.p4_3.into_gpio().into_input().into_adc() };
        // =========================================

        // ========= SERIAL INITIALIZATION =========
        let mut serial =
            Serial::<pac::USART3>::new(device.USART3, uart_tx, uart_rx, 0x41, 0x0, 0x0);
        serial.lcr(
            WLS_A::_8_BIT_CHARACTER_LENGTH,
            SBS_A::_1_STOP_BIT,
            PE_A::DISABLE_PARITY_GENERERATION,
            PS_A::ODD_PARITY,
            BC_A::DISABLED,
            DLAB_A::DISABLED,
        );
        serial.ier(true, false, false, false, false);
        // =========================================

        // ======= UART PLL INITIALIZATION =========
        unsafe {
            uart_pll(&mut device.CGU);
        }
        // =========================================

        // ========= TIMER INITIALIZATION ==========
        let mut timer = Timer0::new_no_pins(device.TIMER0);
        timer.start_timer(false);
        timer.set_match(0, TIMER_PERIOD);
        timer.set_prescaler(1024, Some(0));
        // Safe since the timer is stopped
        unsafe {
            timer.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        timer.start_timer(true);
        // =========================================

        // ========== ADC INITIALIZATION ===========
        let adc: Adc<pac::ADC0> = Adc::<pac::ADC0>::new(device.ADC0, 3);
        // =========================================

        write!(serial, "Initialized.\r\n").unwrap();

        init::LateResources {
            ADC: adc,
            USART: serial,
            TIMER: timer,
            ADC0_IN: adc0_in,
        }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [ADC, TIMER, USART, ADC0_IN])]
    fn TIMER0() {
        // Read fropm the ADC
        write!(resources.USART, "[TIMER0] Reading ADC...").unwrap();
        match resources.ADC.read(resources.ADC0_IN) {
            Ok(v) => { write!(resources.USART, " Done. Value: {}\r\n", v).unwrap(); },
            Err(e) =>{ write!(resources.USART, " Failed. {:?}", e).unwrap(); },
        }

        // Clear interrupt
        resources
            .TIMER
            .clear_interrupt(InterruptType::MatchChannel, 0);
    }
};
