//! # Point-to-point speedtest
//! This program sllows you to strobe the CAN network with a single packet whose
//! starting time is known. Simply run the program, setup your oscilloscope to
//! probe the GPIO output and measure the time taken between rise and fall.
//!
//! At some point, the timer will trigger to send your CAN output, which makes
//! the breakpoint trigger, allowing you to setup single-edge triggering.
//!
//! *REMEMBER: Using debug mode builds will yield _very_ pessimistic results.*

#![feature(panic_info_message)]
#![no_main]
#![no_std]
extern crate panic_halt;

use embedded_hal::digital::v2::OutputPin;
use lpc43xx_hal::{
    pac,
    can::{
        CanBus,
        MsgIface,
        MsgObj,
        IntoCanTx,
        IntoCanRx,
        CanRx,
        CanTx,
    },
    timer::{Timer, InterruptType, EMC0_A, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
    io::pin::{
        split_ports, PinSplit, Pin, IntoGpioPin, Gpio, Output,
        E, U0, U1, U2, U3, U7,
    },
};
use pac::c_can1::cntl::{DAR_A, EIE_A, IE_A, SIE_A};

static TIMER_PERIOD: u32 = 8192;

use rtfm::app;

use heapless::consts::U32;

type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

type Can0Rd = Pin<E, U2, CanRx<pac::C_CAN0>>;
type Can0Td = Pin<E, U3, CanTx<pac::C_CAN0>>;

type Can1Rd = Pin<E, U1, CanRx<pac::C_CAN1>>;
type Can1Td = Pin<E, U0, CanTx<pac::C_CAN1>>;
type OutSignal = Pin<U7, U7, Gpio<Output>>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut SIGNAL: OutSignal = ();
    static mut TIMER: Timer0 = ();
    static mut CAN_0: CanBus<pac::C_CAN0, U32> = ();
    static mut CAN_1: CanBus<pac::C_CAN1, U32> = ();
    static mut RX_OBJ: MsgObj = ();
    static mut TX_OBJ: MsgObj = ();
    static mut TX_STATE: u8 = ();

    #[init]
    fn init() -> init::LateResources {
        let tx_state: u8 = 0;

        // ================= PINS ==================
        let ports = split_ports(device.SCU, device.GPIO_PORT);

        let port_7 = ports
            .port_7
            .split_pins();

        let port_e = ports
            .port_e
            .split_pins();

        let can_0_td: Can0Td = port_e.pe_3.into_can_tx();
        let can_0_rd: Can0Rd = port_e.pe_2.into_can_rx();

        let can_1_td: Can1Td = port_e.pe_0.into_can_tx();
        let can_1_rd: Can1Rd = port_e.pe_1.into_can_rx();
        // =========================================

        // ========= TIMER INITIALIZATION ==========
        let mut timer = Timer0::new_no_pins(device.TIMER0);
        timer.start_timer(false);
        timer.set_match(0, TIMER_PERIOD);
        timer.set_prescaler(1024, Some(0));
        // Safe since the timer is stopped
        unsafe {
            timer.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        timer.start_timer(true);

        // ========= GPIO SETUP  =========
        let mut signal: OutSignal = port_7.p7_7.into_gpio().into_output();
        signal.set_low();

        // ========= CAN 0 INITIALIZATION ==========
        let mut can_0 = CanBus::<pac::C_CAN0, U32>::new(device.C_CAN0, can_0_td, can_0_rd);
        can_0.initialize(
            12_000_000,
            500000,
            IE_A::ENABLE_CAN_INTERRUPTS,
            SIE_A::DISABLE_STATUS_CHANGE_INTERRUPTS,
            EIE_A::DISABLE_ERROR_INTERRUPTS,
            DAR_A::ENABLED,
            false,
        );

        // Object setup
        let tx_obj = MsgObj::new_tx_obj(true, 0x200, 0x8, 0xDEADBEEF, 0x8008135);

        // ========= CAN 1 INITIALIZATION ==========
        let mut can_1 = CanBus::<pac::C_CAN1, U32>::new(device.C_CAN1, can_1_td, can_1_rd);

        can_1.initialize(
            12_000_000,
            500000,
            IE_A::ENABLE_CAN_INTERRUPTS,
            SIE_A::DISABLE_STATUS_CHANGE_INTERRUPTS,
            EIE_A::DISABLE_ERROR_INTERRUPTS,
            DAR_A::ENABLED,
            false,
        );

        let rx_obj = MsgObj::new_rx_obj(true, 0x200, 0x8, 0xfffffff);
        can_1.transfer_obj(MsgIface::Iface1, 1, &rx_obj);

        init::LateResources {
            SIGNAL: signal,
            RX_OBJ: rx_obj,
            TX_OBJ: tx_obj,
            TIMER: timer,
            CAN_0: can_0,
            CAN_1: can_1,
            TX_STATE: tx_state,
        }
    }

    #[interrupt(resources = [SIGNAL, TIMER, CAN_0, TX_OBJ, TX_STATE])]
    fn TIMER0() {
        match resources.TX_STATE {
            0 => resources
                .TX_OBJ
                .msgval(false)
                .mctrl_txrqst(false)
                .cmdmsk_txrqst(false)
                .data_a(true)
                .data_b(true),
            1 => resources
                .TX_OBJ
                .msgval(false)
                .newdat(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            2 => resources
                .TX_OBJ
                .msgval(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            _ => resources
                .TX_OBJ
                .msgval(false)
                .mctrl_txrqst(false)
                .cmdmsk_txrqst(false),
        };

        *resources.TX_STATE += if *resources.TX_STATE < 4 { 1 } else { 0 };
        if *resources.TX_STATE == 2 {
            cortex_m::asm::bkpt();
            resources.SIGNAL.set_high().unwrap();
        }
        resources
            .CAN_0
            .transfer_obj(MsgIface::Iface1, 1, &resources.TX_OBJ);
        resources
            .TIMER
            .clear_interrupt(InterruptType::MatchChannel, 0);

        if *resources.TX_STATE == 3 {
            resources.TIMER.start_timer(false);
        }
    }

    #[interrupt(resources = [TX_OBJ, CAN_0])]
    fn C_CAN0() {
        resources
            .CAN_0
            .can
            .stat
            .modify(|_, w| w.txok().bit(false).rxok().bit(false));

        // Clear newdat
        let mut tx_obj = resources.TX_OBJ;
        tx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
        tx_obj.newdat(false);
        tx_obj.mctrl_txrqst(false);
        tx_obj.cmdmsk_txrqst(false);

        // Write object
        let mut can = resources.CAN_0;
        can.transfer_obj(MsgIface::Iface1, 1, &tx_obj);
    }

    #[interrupt(resources = [CAN_1, SIGNAL, RX_OBJ])]
    fn C_CAN1() {
        (*resources.SIGNAL).set_low();
        resources
            .CAN_1
            .can
            .stat
            .modify(|_, w| w.txok().bit(false).rxok().bit(false));

        // Clear newdat
        let mut rx_obj = resources.RX_OBJ;
        rx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
        rx_obj.newdat(false);
        rx_obj.mctrl_txrqst(false);
        rx_obj.cmdmsk_txrqst(false);

        // Write object
        let can = resources.CAN_1;
        can.transfer_obj(MsgIface::Iface1, 1, &rx_obj);
    }
};
