//! # CAN-bus traffic generator, node B & C
//! Receives and generates a constant stream of high-priority messages on a bus.
#![feature(panic_info_message)]
#![no_main]
#![no_std]
extern crate panic_halt;

use embedded_hal::digital::v2::OutputPin;
use lpc43xx_hal::{
    pac,
    can::{
        CanBus,
        MsgIface,
        MsgObj,
        IntoCanTx,
        IntoCanRx,
        CanRx,
        CanTx,
    },
    io::pin::{
        split_ports, PinSplit, Pin, IntoGpioPin, Gpio, Output,
        E, U0, U1, U2, U3, U7,
    },
    timer::{InterruptType, Timer, EMC0_A, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
};
pub use pac::c_can1::cntl::{CCE_A, DAR_A, EIE_A, IE_A, SIE_A, TEST_A};
use rtfm::app;

use heapless::consts::U32;


type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

type Can0Rd = Pin<E, U2, CanRx<pac::C_CAN0>>;
type Can0Td = Pin<E, U3, CanTx<pac::C_CAN0>>;

type Can1Rd = Pin<E, U1, CanRx<pac::C_CAN1>>;
type Can1Td = Pin<E, U0, CanTx<pac::C_CAN1>>;

type OutSignal = Pin<U7, U7, Gpio<Output>>;

static TIMER_PERIOD: u32 = 8192;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut SIGNAL: OutSignal = ();
    static mut TIMER: Timer0 = ();
    static mut CAN_0: CanBus<pac::C_CAN0, U32> = ();
    static mut CAN_1: CanBus<pac::C_CAN1, U32> = ();
    static mut TX_STATE: u8 = ();
    static mut RX_SYNC_OBJ: MsgObj = ();
    static mut TX_SYNC_OBJ: MsgObj = ();
    static mut RX_B_OBJ: MsgObj = ();
    static mut TX_B_OBJ: MsgObj = ();
    static mut RX_C_OBJ: MsgObj = ();
    static mut TX_C_OBJ: MsgObj = ();

    #[init]
    fn init() -> init::LateResources {
        let tx_state: u8 = 0;

        // ========= TIMER INITIALIZATION ==========
        let mut timer = Timer0::new_no_pins(device.TIMER0);
        timer.start_timer(false);
        timer.set_match(0, TIMER_PERIOD);
        timer.set_prescaler(1024, Some(0));
        unsafe {
            timer.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        timer.start_timer(true);

        // =========== Pins ==============
        let ports = split_ports(device.SCU, device.GPIO_PORT);

        let port_e = ports
            .port_e
            .split_pins();

        let can_0_td: Can0Td = port_e.pe_3.into_can_tx();
        let can_0_rd: Can0Rd = port_e.pe_2.into_can_rx();

        let can_1_td: Can1Td = port_e.pe_0.into_can_tx();
        let can_1_rd: Can1Rd = port_e.pe_1.into_can_rx();

        let port_7 = ports
            .port_7.
            split_pins();
        let mut signal: OutSignal = port_7.p7_7.into_gpio().into_output();
        // ===============================

        // ========= CAN 0 (B) INITIALIZATION ==========
        let mut can_0 = CanBus::<pac::C_CAN0, U32>::new(device.C_CAN0, can_0_td, can_0_rd);

        can_0.initialize(
            12_000_000,
            500000,
            IE_A::ENABLE_CAN_INTERRUPTS,
            SIE_A::DISABLE_STATUS_CHANGE_INTERRUPTS,
            EIE_A::DISABLE_ERROR_INTERRUPTS,
            DAR_A::ENABLED,
            false,
        );

        // Sync setup
        let tx_sync_obj = MsgObj::new_tx_obj(true, 0x400, 0x0, 0x0, 0x0);
        can_0.transfer_obj(MsgIface::Iface1, 1, &tx_sync_obj);

        // Data setup
        let tx_b_obj = MsgObj::new_tx_obj(true, 0x100, 0x8, 0xBBBBBBBB, 0xC2C2D3D3);
        can_0.transfer_obj(MsgIface::Iface1, 2, &tx_b_obj);
        let rx_b_obj = MsgObj::new_rx_obj(true, 0x800, 0x8, 0xffffffff);
        can_0.transfer_obj(MsgIface::Iface1, 3, &rx_b_obj);

        // ========= CAN 1 (C) INITIALIZATION ==========
        let mut can_1 = CanBus::<pac::C_CAN1, U32>::new(device.C_CAN1, can_1_td, can_1_rd);

        can_1.initialize(
            12_000_000,
            500000,
            IE_A::ENABLE_CAN_INTERRUPTS,
            SIE_A::DISABLE_STATUS_CHANGE_INTERRUPTS,
            EIE_A::DISABLE_ERROR_INTERRUPTS,
            DAR_A::ENABLED,
            false,
        );

        // Sync message setup
        let rx_sync_obj = MsgObj::new_rx_obj(true, 0x400, 0x0, 0xfffffff);
        can_1.transfer_obj(MsgIface::Iface1, 1, &rx_sync_obj);

        // Data setup
        let tx_c_obj = MsgObj::new_tx_obj(true, 0x800, 0x8, 0xCCCCCCCC, 0xE4E4F5F5);
        can_1.transfer_obj(MsgIface::Iface1, 2, &tx_c_obj);
        let rx_c_obj = MsgObj::new_rx_obj(true, 0x100, 0x8, 0xffffffff);
        can_1.transfer_obj(MsgIface::Iface1, 3, &rx_c_obj);

        cortex_m::asm::bkpt();

        init::LateResources {
            SIGNAL: signal,
            TIMER: timer,
            CAN_0: can_0,
            CAN_1: can_1,
            TX_STATE: tx_state,
            RX_SYNC_OBJ: rx_sync_obj,
            TX_SYNC_OBJ: tx_sync_obj,
            RX_B_OBJ: rx_b_obj,
            TX_B_OBJ: tx_b_obj,
            RX_C_OBJ: rx_c_obj,
            TX_C_OBJ: tx_c_obj,
        }
    }

    #[interrupt(resources = [SIGNAL, TIMER, CAN_0, CAN_1, TX_SYNC_OBJ, TX_B_OBJ, TX_C_OBJ, TX_STATE])]
    fn TIMER0() {
        // Setup SYNC transfer
        match resources.TX_STATE {
            28 => resources
                .TX_SYNC_OBJ
                .msgval(false)
                .mctrl_txrqst(false)
                .cmdmsk_txrqst(false),
            29 => resources
                .TX_SYNC_OBJ
                .msgval(false)
                .newdat(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            30 => {
                cortex_m::asm::bkpt();
                resources.SIGNAL.set_high().unwrap();
                resources
                    .TX_SYNC_OBJ
                    .msgval(true)
                    .mctrl_txrqst(true)
                    .cmdmsk_txrqst(true)
            }
            31 => resources
                .TX_SYNC_OBJ
                .msgval(false)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            _ => resources.TX_SYNC_OBJ,
        };

        // Setup CAN0 transfers
        match resources.TX_STATE {
            0 => resources
                .TX_B_OBJ
                .msgval(false)
                .mctrl_txrqst(false)
                .cmdmsk_txrqst(false)
                .data_a(true)
                .data_b(true),
            1 => resources
                .TX_B_OBJ
                .msgval(false)
                .newdat(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            _ => resources
                .TX_B_OBJ
                .msgval(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
        };

        // Setup CAN1 transfers
        match resources.TX_STATE {
            0 => resources
                .TX_C_OBJ
                .msgval(false)
                .mctrl_txrqst(false)
                .cmdmsk_txrqst(false)
                .data_a(true)
                .data_b(true),
            1 => resources
                .TX_C_OBJ
                .msgval(false)
                .newdat(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
            _ => resources
                .TX_C_OBJ
                .msgval(true)
                .mctrl_txrqst(true)
                .cmdmsk_txrqst(true),
        };

        *resources.TX_STATE = (*resources.TX_STATE + 1) % 64;

        // Node B transmissions
        resources
            .CAN_0
            .transfer_obj(MsgIface::Iface1, 1, &resources.TX_SYNC_OBJ);
        resources
            .CAN_0
            .transfer_obj(MsgIface::Iface1, 2, &resources.TX_B_OBJ);

        // Node C transmissions
        resources
            .CAN_1
            .transfer_obj(MsgIface::Iface1, 2, &resources.TX_C_OBJ);

        resources
            .TIMER
            .clear_interrupt(InterruptType::MatchChannel, 0);
    }

    #[interrupt(resources = [TX_B_OBJ, RX_B_OBJ, CAN_0])]
    fn C_CAN0() {
        let mut can = resources.CAN_0;
        let intid = can.can.int.read().intid15_0().bits();

        match intid {
            2 => {
                can.can
                    .stat
                    .modify(|_, w| w.txok().bit(false).rxok().bit(false));

                // Set newdat
                let mut tx_obj = resources.TX_B_OBJ;
                tx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
                tx_obj.newdat(false);
                tx_obj.mctrl_txrqst(false);
                tx_obj.cmdmsk_txrqst(false);

                // Write object
                can.transfer_obj(MsgIface::Iface1, 2, &tx_obj);
            }
            3 => {
                can.can
                    .stat
                    .modify(|_, w| w.txok().bit(false).rxok().bit(false));

                // Clear newdat
                let mut rx_obj = resources.RX_B_OBJ;
                rx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
                rx_obj.newdat(false);
                rx_obj.mctrl_txrqst(false);
                rx_obj.cmdmsk_txrqst(false);

                // Write object
                can.transfer_obj(MsgIface::Iface1, 3, &rx_obj);
            }
            _ => (),
        }
    }

    #[interrupt(resources = [TX_C_OBJ, RX_C_OBJ, RX_SYNC_OBJ, CAN_1, SIGNAL])]
    fn C_CAN1() {
        let mut can = resources.CAN_1;
        let intid = can.can.int.read().intid15_0().bits();

        match intid {
            1 => {
                // Received sync, toggle and break
                resources.SIGNAL.set_low().unwrap();
                can.can
                    .stat
                    .modify(|_, w| w.txok().bit(false).rxok().bit(false));

                // Clear newdat
                let mut rx_obj = resources.RX_SYNC_OBJ;
                rx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, CLRINTPND, WR_RD
                rx_obj.newdat(false);
                rx_obj.mctrl_txrqst(false);
                rx_obj.cmdmsk_txrqst(false);

                // Write object
                can.transfer_obj(MsgIface::Iface1, 1, &rx_obj);

                // The transfer has been made, pause
                // cortex_m::asm::bkpt();
            }
            2 => {
                can.can
                    .stat
                    .modify(|_, w| w.txok().bit(false).rxok().bit(false));

                // Set newdat
                let mut tx_obj = resources.TX_C_OBJ;
                tx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
                tx_obj.newdat(false);
                tx_obj.mctrl_txrqst(false);
                tx_obj.cmdmsk_txrqst(false);

                // Write object
                can.transfer_obj(MsgIface::Iface1, 2, &tx_obj);
            }
            3 => {
                can.can
                    .stat
                    .modify(|_, w| w.txok().bit(false).rxok().bit(false));

                // Clear newdat
                let mut rx_obj = resources.RX_C_OBJ;
                rx_obj.cmdmsk = (1 << 2) | (1 << 3) | (1 << 7); // NEWDAT, WR_RD
                rx_obj.newdat(false);
                rx_obj.mctrl_txrqst(false);
                rx_obj.cmdmsk_txrqst(false);

                // Write object
                can.transfer_obj(MsgIface::Iface1, 3, &rx_obj);
            }
            _ => (),
        }
    }
};
