#![feature(panic_info_message)]
#![no_main]
#![no_std]

use rtfm::app;

// UART driver
use lpc43xx_hal::{
    io::pin::{split_ports, PinSplit},
    pac,
    serial::Serial,
    timer::{Timer, EMC0_A, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
};
use pac::usart0::lcr::{BC_A, DLAB_A, PE_A, PS_A, SBS_A, WLS_A};

static TIMER_PERIOD: u32 = 8192;

// Panic handler for USART3
use lpc43xx_hal::{debug::PanicInfo, uart_panic_handler};
uart_panic_handler!(pac::USART3);

type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut TIMER: Timer0 = ();

    #[init]
    fn init() -> init::LateResources {
        // Pins
        let port_9 = split_ports(device.SCU, device.GPIO_PORT)
            .port_9
            .split_pins();
        let uart_tx = port_9.p9_3;
        let uart_rx = port_9.p9_4;

        let mut serial =
            Serial::<pac::USART3>::new(device.USART3, uart_tx, uart_rx, 0x41, 0x0, 0x0);
        serial.lcr(
            WLS_A::_8_BIT_CHARACTER_LENGTH,
            SBS_A::_1_STOP_BIT,
            PE_A::DISABLE_PARITY_GENERERATION,
            PS_A::ODD_PARITY,
            BC_A::DISABLED,
            DLAB_A::DISABLED,
        );
        serial.ier(true, false, false, false, false);

        // Configure timer
        let mut timer = Timer0::new_no_pins(device.TIMER0);
        timer.start_timer(false);
        timer.set_match(0, TIMER_PERIOD);
        timer.set_prescaler(1024, Some(0));
        // Safe since the timer is stopped
        unsafe {
            timer.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        timer.start_timer(true);

        // Configure PLL1 for use with UART3
        unsafe {
            // Select the IRC as BASE_M4_CLK source
            device.CGU.base_m4_clk.write(|w| w.clk_sel().bits(0x1));

            // Enable the crystal oscillator
            device.CGU.xtal_osc_ctrl.modify(|_, w| w.hf().bit(false));
            device
                .CGU
                .xtal_osc_ctrl
                .modify(|_, w| w.enable().bit(false));

            // Wait 250us
            for i in 0..3_000 {
                core::ptr::read_volatile(&i);
            }

            device.CGU.pll1_ctrl.write(|w| {
                w.pd()
                    .bit(false)
                    .clk_sel()
                    .bits(0x6) // Select the crystal oscillator as clock source
                    .msel()
                    .bits(20)
                    .nsel()
                    .bits(1) // Select the M and N divider values
                    .psel()
                    .bits(0)
                    .fbsel()
                    .bit(true)
                    .bypass()
                    .bit(false)
                    .direct()
                    .bit(false)
            });

            // Wait for the PLL to lock
            while !device.CGU.pll1_stat.read().lock().bit() {}

            // Set PLL1 p-divider to divide by 2 (DIRECT=0; PSEL=0)
            device
                .CGU
                .pll1_ctrl
                .modify(|_, w| w.direct().bit(false).psel().bits(0x0));

            // Select PLL1 as base_m4_clk source
            device.CGU.base_m4_clk.modify(|_, w| w.clk_sel().bits(0x09));

            // Wait for 50us
            for i in 0..600 {
                core::ptr::read_volatile(&i);
            }

            // Enable direct output mode
            device.CGU.pll1_ctrl.modify(|_, w| w.direct().bit(true));

            // Enable clock on UART3
            device.CGU.base_uart3_clk.write(|w| w.clk_sel().bits(0x09)); // configure crystal oscillator
        }

        unsafe {
            PANIC_UART = Some(serial);
        }
        init::LateResources { TIMER: timer }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [TIMER])]
    fn TIMER0() {
        panic!("this is being printed over UART");
    }
};
