#![no_main]
#![no_std]
#[allow(unused_extern_crates)]
extern crate panic_halt;

use lpc43xx_hal::{
    creg::creg6,
    pac, sct,
    timer::{Timer, EMC0_A, InterruptType, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
};
use rtfm::app;

static TIMER_PERIOD: u32 = 0x100;
static PWM_PERIOD: u32 = 0xffff;

type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut TIMER: Timer0 = ();
    static mut SCT: pac::SCT = ();
    static mut PHASE: u16 = 0;

    #[init]
    fn init() -> init::LateResources {
        // Setup SCT clock and I/O
        creg6::control_ctout(&mut device.CREG, creg6::CTOUTCTRL_A::SCT_OUTPUTS_ONLY);

        // Configure P7_7 as CTOUT_8 (mode=0x1)
        device.SCU.sfsp7_7.write(|w| w.mode().function_1());

        // Configure timer
        let mut t = Timer0::new_no_pins(device.TIMER0);
        t.start_timer(false);
        t.set_match(0, TIMER_PERIOD);
        t.set_prescaler(1024, Some(0));
        // Safe since the timer is stopped
        unsafe {
            t.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        t.start_timer(true);

        // Configure SCT
        sct::config::setup(
            &mut device.SCT,
            sct::config::UNIFY_A::_16_BIT,
            sct::config::CLKMODE_A::BUS_CLOCK,
            sct::config::CKSEL_A::RISING_EDGES_ON_INPUT_0,
            false,
            false,
            0x0,
        );
        sct::config::control(
            &mut device.SCT,
            sct::util::CtrlBlock::High,
            false,
            false,
            false,
            false,
            Some(0x0),
        );

        // Configure reset timer
        device.SCT.ctrl.modify(|_, w| w.stop_h().bit(true));

        // MR0 config
        sct::config::set_match_reg(&mut device.SCT, sct::util::CtrlBlock::High, 0, PWM_PERIOD);
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::High, 0, PWM_PERIOD);

        // Set limit, state mask and enable event
        sct::config::set_limit(&mut device.SCT, sct::util::CtrlBlock::High, 1 << 0);
        sct::set_state_mask(&device.SCT.ev0_state, 0x1);
        sct::event::config_match_event(
            &device.SCT.ev0_ctrl,
            sct::event::HEVENT_A::H_STATE,
            sct::event::COMBMODE_A::MATCH,
            0,
        );
        sct::ctrl::start_timer(&mut device.SCT, sct::util::CtrlBlock::High, 0);

        // MR1 config
        sct::config::set_match_reg(
            &mut device.SCT,
            sct::util::CtrlBlock::High,
            1,
            PWM_PERIOD / 2,
        );
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::High, 1, PWM_PERIOD / 2);

        // Set register mode and event action
        sct::config::set_register_mode(&mut device.SCT, sct::util::CtrlBlock::High, 0 << 1);
        sct::event::config_match_event(
            &device.SCT.ev1_ctrl,
            sct::event::HEVENT_A::H_STATE,
            sct::event::COMBMODE_A::MATCH,
            1,
        );

        // Set output on event 8
        sct::event::set_event_set(&device.SCT.out8_set, 0);
        sct::event::set_event_clear(&device.SCT.out8_clr, 1);

        sct::set_state_mask(&device.SCT.ev1_state, 0x1);
        sct::ctrl::start_timer(&mut device.SCT, sct::util::CtrlBlock::High, 1);

        device.SCT.ctrl.modify(|_, w| w.stop_h().bit(false));

        init::LateResources {
            TIMER: t,
            SCT: device.SCT,
        }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [SCT, TIMER, PHASE])]
    fn TIMER0() {
        *resources.PHASE = (*resources.PHASE + 1) % 360;
        let intensity = lut_sin(*resources.PHASE);
        sct::config::set_reload(
            &resources.SCT,
            sct::util::CtrlBlock::High,
            1,
            intensity as u32,
        );
        resources
            .TIMER
            .clear_interrupt(InterruptType::MatchChannel, 0);
    }
};

fn lut_sin(x: u16) -> u16 {
    match x {
        0 => 0x7fff,
        1 => 0x823b,
        2 => 0x8477,
        3 => 0x86b2,
        4 => 0x88ed,
        5 => 0x8b27,
        6 => 0x8d60,
        7 => 0x8f98,
        8 => 0x91cf,
        9 => 0x9405,
        10 => 0x9639,
        11 => 0x986b,
        12 => 0x9a9c,
        13 => 0x9cca,
        14 => 0x9ef6,
        15 => 0xa120,
        16 => 0xa347,
        17 => 0xa56b,
        18 => 0xa78d,
        19 => 0xa9ab,
        20 => 0xabc6,
        21 => 0xadde,
        22 => 0xaff2,
        23 => 0xb202,
        24 => 0xb40f,
        25 => 0xb617,
        26 => 0xb81b,
        27 => 0xba1b,
        28 => 0xbc16,
        29 => 0xbe0d,
        30 => 0xbfff,
        31 => 0xc1ec,
        32 => 0xc3d3,
        33 => 0xc5b5,
        34 => 0xc792,
        35 => 0xc96a,
        36 => 0xcb3b,
        37 => 0xcd07,
        38 => 0xcecd,
        39 => 0xd08c,
        40 => 0xd246,
        41 => 0xd3f8,
        42 => 0xd5a5,
        43 => 0xd74a,
        44 => 0xd8e9,
        45 => 0xda81,
        46 => 0xdc12,
        47 => 0xdd9c,
        48 => 0xdf1e,
        49 => 0xe099,
        50 => 0xe20c,
        51 => 0xe378,
        52 => 0xe4dc,
        53 => 0xe638,
        54 => 0xe78c,
        55 => 0xe8d9,
        56 => 0xea1c,
        57 => 0xeb58,
        58 => 0xec8b,
        59 => 0xedb6,
        60 => 0xeed8,
        61 => 0xeff2,
        62 => 0xf103,
        63 => 0xf20b,
        64 => 0xf30a,
        65 => 0xf400,
        66 => 0xf4ee,
        67 => 0xf5d2,
        68 => 0xf6ac,
        69 => 0xf77e,
        70 => 0xf846,
        71 => 0xf905,
        72 => 0xf9bb,
        73 => 0xfa67,
        74 => 0xfb09,
        75 => 0xfba2,
        76 => 0xfc31,
        77 => 0xfcb7,
        78 => 0xfd32,
        79 => 0xfda4,
        80 => 0xfe0d,
        81 => 0xfe6b,
        82 => 0xfec0,
        83 => 0xff0a,
        84 => 0xff4b,
        85 => 0xff82,
        86 => 0xffaf,
        87 => 0xffd2,
        88 => 0xffeb,
        89 => 0xfffa,
        90 => 0xffff,
        91 => 0xfffa,
        92 => 0xffeb,
        93 => 0xffd2,
        94 => 0xffaf,
        95 => 0xff82,
        96 => 0xff4b,
        97 => 0xff0a,
        98 => 0xfec0,
        99 => 0xfe6b,
        100 => 0xfe0d,
        101 => 0xfda4,
        102 => 0xfd32,
        103 => 0xfcb7,
        104 => 0xfc31,
        105 => 0xfba2,
        106 => 0xfb09,
        107 => 0xfa67,
        108 => 0xf9bb,
        109 => 0xf905,
        110 => 0xf846,
        111 => 0xf77e,
        112 => 0xf6ac,
        113 => 0xf5d2,
        114 => 0xf4ee,
        115 => 0xf400,
        116 => 0xf30a,
        117 => 0xf20b,
        118 => 0xf103,
        119 => 0xeff2,
        120 => 0xeed8,
        121 => 0xedb6,
        122 => 0xec8b,
        123 => 0xeb58,
        124 => 0xea1c,
        125 => 0xe8d9,
        126 => 0xe78c,
        127 => 0xe638,
        128 => 0xe4dc,
        129 => 0xe378,
        130 => 0xe20c,
        131 => 0xe099,
        132 => 0xdf1e,
        133 => 0xdd9c,
        134 => 0xdc12,
        135 => 0xda81,
        136 => 0xd8e9,
        137 => 0xd74a,
        138 => 0xd5a5,
        139 => 0xd3f8,
        140 => 0xd246,
        141 => 0xd08c,
        142 => 0xcecd,
        143 => 0xcd07,
        144 => 0xcb3b,
        145 => 0xc96a,
        146 => 0xc792,
        147 => 0xc5b5,
        148 => 0xc3d3,
        149 => 0xc1ec,
        150 => 0xbfff,
        151 => 0xbe0d,
        152 => 0xbc16,
        153 => 0xba1b,
        154 => 0xb81b,
        155 => 0xb617,
        156 => 0xb40f,
        157 => 0xb202,
        158 => 0xaff2,
        159 => 0xadde,
        160 => 0xabc6,
        161 => 0xa9ab,
        162 => 0xa78d,
        163 => 0xa56b,
        164 => 0xa347,
        165 => 0xa120,
        166 => 0x9ef6,
        167 => 0x9cca,
        168 => 0x9a9c,
        169 => 0x986b,
        170 => 0x9639,
        171 => 0x9405,
        172 => 0x91cf,
        173 => 0x8f98,
        174 => 0x8d60,
        175 => 0x8b27,
        176 => 0x88ed,
        177 => 0x86b2,
        178 => 0x8477,
        179 => 0x823b,
        180 => 0x7fff,
        181 => 0x7dc3,
        182 => 0x7b87,
        183 => 0x794c,
        184 => 0x7711,
        185 => 0x74d7,
        186 => 0x729e,
        187 => 0x7066,
        188 => 0x6e2f,
        189 => 0x6bf9,
        190 => 0x69c5,
        191 => 0x6793,
        192 => 0x6562,
        193 => 0x6334,
        194 => 0x6108,
        195 => 0x5ede,
        196 => 0x5cb7,
        197 => 0x5a93,
        198 => 0x5871,
        199 => 0x5653,
        200 => 0x5438,
        201 => 0x5220,
        202 => 0x500c,
        203 => 0x4dfc,
        204 => 0x4bef,
        205 => 0x49e7,
        206 => 0x47e3,
        207 => 0x45e3,
        208 => 0x43e8,
        209 => 0x41f1,
        210 => 0x3fff,
        211 => 0x3e12,
        212 => 0x3c2b,
        213 => 0x3a49,
        214 => 0x386c,
        215 => 0x3694,
        216 => 0x34c3,
        217 => 0x32f7,
        218 => 0x3131,
        219 => 0x2f72,
        220 => 0x2db8,
        221 => 0x2c06,
        222 => 0x2a59,
        223 => 0x28b4,
        224 => 0x2715,
        225 => 0x257d,
        226 => 0x23ec,
        227 => 0x2262,
        228 => 0x20e0,
        229 => 0x1f65,
        230 => 0x1df2,
        231 => 0x1c86,
        232 => 0x1b22,
        233 => 0x19c6,
        234 => 0x1872,
        235 => 0x1725,
        236 => 0x15e2,
        237 => 0x14a6,
        238 => 0x1373,
        239 => 0x1248,
        240 => 0x1126,
        241 => 0x100c,
        242 => 0xefb,
        243 => 0xdf3,
        244 => 0xcf4,
        245 => 0xbfe,
        246 => 0xb10,
        247 => 0xa2c,
        248 => 0x952,
        249 => 0x880,
        250 => 0x7b8,
        251 => 0x6f9,
        252 => 0x643,
        253 => 0x597,
        254 => 0x4f5,
        255 => 0x45c,
        256 => 0x3cd,
        257 => 0x347,
        258 => 0x2cc,
        259 => 0x25a,
        260 => 0x1f1,
        261 => 0x193,
        262 => 0x13e,
        263 => 0xf4,
        264 => 0xb3,
        265 => 0x7c,
        266 => 0x4f,
        267 => 0x2c,
        268 => 0x13,
        269 => 0x4,
        270 => 0x0,
        271 => 0x4,
        272 => 0x13,
        273 => 0x2c,
        274 => 0x4f,
        275 => 0x7c,
        276 => 0xb3,
        277 => 0xf4,
        278 => 0x13e,
        279 => 0x193,
        280 => 0x1f1,
        281 => 0x25a,
        282 => 0x2cc,
        283 => 0x347,
        284 => 0x3cd,
        285 => 0x45c,
        286 => 0x4f5,
        287 => 0x597,
        288 => 0x643,
        289 => 0x6f9,
        290 => 0x7b8,
        291 => 0x880,
        292 => 0x952,
        293 => 0xa2c,
        294 => 0xb10,
        295 => 0xbfe,
        296 => 0xcf4,
        297 => 0xdf3,
        298 => 0xefb,
        299 => 0x100c,
        300 => 0x1126,
        301 => 0x1248,
        302 => 0x1373,
        303 => 0x14a6,
        304 => 0x15e2,
        305 => 0x1725,
        306 => 0x1872,
        307 => 0x19c6,
        308 => 0x1b22,
        309 => 0x1c86,
        310 => 0x1df2,
        311 => 0x1f65,
        312 => 0x20e0,
        313 => 0x2262,
        314 => 0x23ec,
        315 => 0x257d,
        316 => 0x2715,
        317 => 0x28b4,
        318 => 0x2a59,
        319 => 0x2c06,
        320 => 0x2db8,
        321 => 0x2f72,
        322 => 0x3131,
        323 => 0x32f7,
        324 => 0x34c3,
        325 => 0x3694,
        326 => 0x386c,
        327 => 0x3a49,
        328 => 0x3c2b,
        329 => 0x3e12,
        330 => 0x3fff,
        331 => 0x41f1,
        332 => 0x43e8,
        333 => 0x45e3,
        334 => 0x47e3,
        335 => 0x49e7,
        336 => 0x4bef,
        337 => 0x4dfc,
        338 => 0x500c,
        339 => 0x5220,
        340 => 0x5438,
        341 => 0x5653,
        342 => 0x5871,
        343 => 0x5a93,
        344 => 0x5cb7,
        345 => 0x5ede,
        346 => 0x6108,
        347 => 0x6334,
        348 => 0x6562,
        349 => 0x6793,
        350 => 0x69c5,
        351 => 0x6bf9,
        352 => 0x6e2f,
        353 => 0x7066,
        354 => 0x729e,
        355 => 0x74d7,
        356 => 0x7711,
        357 => 0x794c,
        358 => 0x7b87,
        359 => 0x7dc3,
        _ => 0,
    }
}
