// Example of sct in 16-bit mode.
// Both the high and low counters are used to make pwm's with differant periods.
// Implements a slow PWM output on P1_11 (low counter to CTOUT_15) and P1_7 (higher counter to CTOUT_13).
// match0_l and match1_l is used for the low counter.
// match0_h and match1_h is used for the high counter.
// Each counter has 2 events
// PWM duty is changed in timer interrupt.
// get_reload() is tested by reading back the reload registers. 
#![no_main]
#![no_std]
#[allow(unused_extern_crates)]
extern crate panic_halt;

use lpc43xx_hal::{
    creg::creg6,
    pac, sct,
    timer::{Timer, EMC0_A, InterruptType, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
};
use rtfm::app;

static TIMER_PERIOD: u32 = 0xf000;
static PWM_PERIOD_L: u32 = 0x4080;
static PWM_PERIOD_H: u32 = 0xc080;

type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut TIMER: Timer0 = ();
    static mut SCT: pac::SCT = ();
    static mut PHASE: u16 = 0;

    #[init]
    fn init() -> init::LateResources {
        // Setup SCT clock and I/O
        creg6::control_ctout(&mut device.CREG, creg6::CTOUTCTRL_A::SCT_OUTPUTS_ONLY);

        // P1_11 as CTOUT_15 
        device.SCU.sfsp1_11.write(|w| w.mode().function_2());

        // P1_7 as CTOUT_13 
        device.SCU.sfsp1_7.write(|w| w.mode().function_2());

        // Configure timer
        let mut t = Timer0::new_no_pins(device.TIMER0);
        t.start_timer(false);
        t.set_match(0, TIMER_PERIOD);
        t.set_prescaler(1024, Some(0));
        // Safe since the timer is stopped
        unsafe {
            t.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        t.start_timer(true);

        // Configure SCT
        sct::config::setup(
            &mut device.SCT,
            sct::config::UNIFY_A::_16_BIT,
            sct::config::CLKMODE_A::BUS_CLOCK,
            sct::config::CKSEL_A::RISING_EDGES_ON_INPUT_0,
            false,
            false,
            0x0,
        );
        sct::config::control(
            &mut device.SCT,
            sct::util::CtrlBlock::Low,          
            false,
            false,
            false,
            false,
            Some(0x0),
        );
        
        sct::config::control(
            &mut device.SCT,
            sct::util::CtrlBlock::High,          
            false,
            false,
            false,
            false,
            Some(0x0),
        );
        
        // Configure reset timer
        device.SCT.ctrl.modify(|_, w| w.stop_h().bit(true));
        device.SCT.ctrl.modify(|_, w| w.stop_l().bit(true));

        // MR0 config Low limit
        sct::config::set_match_reg(&mut device.SCT, sct::util::CtrlBlock::Low, 0, PWM_PERIOD_L);
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::Low, 0, PWM_PERIOD_L);
                
        // Set limit, state mask and enable event
        sct::config::set_limit(&mut device.SCT, sct::util::CtrlBlock::Low, 1 << 0);
        sct::set_state_mask(&device.SCT.ev0_state, 0x1);
        sct::event::config_match_event(
            &device.SCT.ev0_ctrl,
            sct::event::HEVENT_A::L_STATE,      
            sct::event::COMBMODE_A::MATCH,
            0,
        );

        // MR2 config High limit/wrap
        sct::config::set_match_reg(&mut device.SCT, sct::util::CtrlBlock::High, 0, PWM_PERIOD_H);
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::High, 0, PWM_PERIOD_H);
                
        // Set limit, state mask and enable event
        sct::config::set_limit(&mut device.SCT, sct::util::CtrlBlock::High, 1 << 2);
        sct::set_state_mask(&device.SCT.ev2_state, 0x1);
        sct::event::config_match_event(
            &device.SCT.ev2_ctrl,
            sct::event::HEVENT_A::H_STATE,      
            sct::event::COMBMODE_A::MATCH,
            0,
        );


        // MR1 config Low toggle
        sct::config::set_match_reg(
            &mut device.SCT,
            sct::util::CtrlBlock::Low,
            1,
            PWM_PERIOD_L / 2,
        );
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::Low, 1, PWM_PERIOD_L / 2);

        // Set register mode and event action
        sct::config::set_register_mode(&mut device.SCT, sct::util::CtrlBlock::Low, 0 << 1);
        sct::event::config_match_event(
            &device.SCT.ev1_ctrl,
            sct::event::HEVENT_A::L_STATE,
            sct::event::COMBMODE_A::MATCH,
            1,
        );
        sct::set_state_mask(&device.SCT.ev1_state, 0x1);
  


        // MR3 config High toggle
        sct::config::set_match_reg(
            &mut device.SCT,
            sct::util::CtrlBlock::High,
            1,
            PWM_PERIOD_H / 2,
        );
        sct::config::set_reload(&device.SCT, sct::util::CtrlBlock::High, 1, PWM_PERIOD_H / 2);

        // Set register mode and event action
        sct::config::set_register_mode(&mut device.SCT, sct::util::CtrlBlock::High, 0 << 3);
        sct::event::config_match_event(
            &device.SCT.ev3_ctrl,
            sct::event::HEVENT_A::H_STATE,
            sct::event::COMBMODE_A::MATCH,
            1,
        );

        sct::set_state_mask(&device.SCT.ev3_state, 0x1);

        // Set output on event
        sct::event::set_event_set(&device.SCT.out15_set, 0);
        sct::event::set_event_clear(&device.SCT.out15_clr, 1);

        sct::event::set_event_set(&device.SCT.out13_set, 2);
        sct::event::set_event_clear(&device.SCT.out13_clr, 3);

        device.SCT.ctrl.modify(|_, w| w.stop_l().bit(false));
        device.SCT.ctrl.modify(|_, w| w.stop_h().bit(false));

        init::LateResources {
            TIMER: t,
            SCT: device.SCT,
        }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [SCT, TIMER, PHASE])]
    fn TIMER0() {
        let _get_reload_l: u32 = sct::config::get_reload(
            &resources.SCT,
            sct::util::CtrlBlock::Low,
            1,
        );
        let _get_reload_h: u32 = sct::config::get_reload(
            &resources.SCT,
            sct::util::CtrlBlock::High,
            1,
        );
        
        *resources.PHASE = (*resources.PHASE + 1) % 4;
        let intensity = PWM_PERIOD_L/(*resources.PHASE as u32 + 2);
        
        sct::config::set_reload(
            &resources.SCT,
            sct::util::CtrlBlock::Low,
            1,
            intensity as u32,
        );
        sct::config::set_reload(
            &resources.SCT,
            sct::util::CtrlBlock::High,
            1,
            intensity as u32 + 0xf,
        );
        
        resources
            .TIMER
            .clear_interrupt(InterruptType::MatchChannel, 0);
    }
};
