//! Uses the SPI abstractions in an RTFM application.
//!
//! SSP1 is configured as a SPI master and the SPI peripheral is configured as a SPI slave.
//! The pin mapping is for the Open4352C board.
//! Connect (PF_5, P3_3), (PF_7, P3_6), (P1_5, P3_7), (PF_6, P3_8)
//!
//! The master has two streams of communication: one that counts up from 1 to 10 and one that
//! counts down from 99 to 90. Then they restart.
//!
//! The slave simply echoes back what the master sent.
//!
//! Run in a debugger and observe the `rx` variable when the breakpoint in the master interrupt is
//! reached.
//!
//! Expected
//! ```
//! (Discard, ??)
//! (CountUp, 1)
//! (CountDown, 99)
//! (CountUp, 2)
//! (CountDown, 98)
//! (CountUp, 3)
//! (CountDown, 97)
//! (CountUp, 4)
//! (CountDown, 96)
//! (CountUp, 5)
//! (CountDown, 95)
//! (CountUp, 6)
//! (CountDown, 94)
//! (CountUp, 7)
//! (CountDown, 93)
//! (CountUp, 8)
//! (CountDown, 92)
//! (CountUp, 9)
//! (CountDown, 91)
//! (CountUp, 10)
//! (CountDown, 90)
//! (CountUp, 1)
//! (CountDown, 99)
//! (CountUp, 2)
//! ...
//! ```

#![no_main]
#![no_std]

extern crate panic_halt;

use rtfm::app;

use driver::pac;
use embedded_hal::spi::{FullDuplex, MODE_0};
use lpc43xx_hal as driver;

pub enum SpiSlaveRead {
    Echo,
}

pub enum SpiMasterRead {
    Discard,
    CountUp,
    CountDown,
}

type Spi0 = driver::spi::spi::Spi;
type Spi1 = driver::spi::Spi<pac::SSP1, SpiMasterRead>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut SPI_MASTER: Spi1 = ();
    static mut SPI_SLAVE: Spi0 = ();

    #[init]
    fn init() -> init::LateResources {
        // SSP1 (master) pin configuration
        // SSP1_SCK
        device
            .SCU
            .sfspf_4
            .write(|w| w.mode().function_0_default().epun().enable_pull_up());
        // SSP1_MISO
        device
            .SCU
            .sfspf_6
            .write(|w| w.mode().function_2().ezi().enable_input_buffer());
        // SSP1_MOSI
        device.SCU.sfsp1_4.write(|w| w.mode().function_5());
        // SSP1_SSEL
        device.SCU.sfspf_5.write(|w| w.mode().function_2());

        // SPI (slave) pin configuration
        // SPI_SCK
        device
            .SCU
            .sfsp3_3
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());
        // SPI_MISO
        device.SCU.sfsp3_6.write(|w| w.mode().function_1());
        // SPI_MOSI
        device
            .SCU
            .sfsp3_7
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());
        // SPI_SSEL
        device
            .SCU
            .sfsp3_8
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());

        let spi_slave = Spi0::new(device.SPI, MODE_0, false, true);

        // Discard the first response as we can't expect the slave to echo before we have sent
        // anything
        let spi_master = Spi1::new(device.SSP1, MODE_0, SpiMasterRead::Discard);
        init::LateResources {
            SPI_MASTER: spi_master,
            SPI_SLAVE: spi_slave,
        }
    }

    #[idle(resources = [SPI_MASTER])]
    fn idle() -> ! {
        resources.SPI_MASTER.lock(|spi| {
            spi.transact(1, SpiMasterRead::CountUp)
                .expect("initial idle send fail");
            spi.transact(99, SpiMasterRead::CountDown)
                .expect("initial idle send fail");
        });

        loop {
            cortex_m::asm::wfi();
        }
    }

    #[interrupt(priority = 2, resources = [SPI_SLAVE])]
    fn SPI_INT() {
        let spi = resources.SPI_SLAVE;
        // ideally you would handle the errors too
        let r = spi.read().unwrap();
        spi.send(r).unwrap();
    }

    #[interrupt(resources = [SPI_MASTER])]
    fn SSP1() {
        resources.SPI_MASTER.lock(|spi| {
            while spi.receive_not_empty() {
                let rx = spi.spi_irq().expect("master read fail");
                cortex_m::asm::bkpt();
                match rx {
                    (SpiMasterRead::Discard, _) => (),
                    (SpiMasterRead::CountUp, r) => {
                        let r = if r < 10 { r + 1 } else { 1 };
                        spi.transact(r, SpiMasterRead::CountUp)
                            .expect("master send fail");
                    }
                    (SpiMasterRead::CountDown, r) => {
                        let r = if r > 90 { r - 1 } else { 99 };

                        spi.transact(r, SpiMasterRead::CountDown)
                            .expect("master send fail");
                    }
                };
            }
        });
    }
};
