//! SSP1 is configured as a SPI master and the SPI peripheral is configured as a SPI slave.
//! The pin mapping is for the Open4352C board.
//! Connect (PF_5, P3_3), (PF_7, P3_6), (P1_5, P3_7), (PF_6, P3_8)
//!
//! The master sends two values followed by one garbage byte to read the second response.
//! The slave responds with the last seen value plus 10.
//! The master adds one to the responses and then it prepares to send the new values.
//! Initially, (0, 100, 200) is sent. The response to the last message will be ignored.
//!
//! Run in a debugger and observe the `buf` variable when the breakpoint in the master interrupt is
//! reached.
//!
//! Expected
//! ```
//! [0, 100, 200]
//! [11, 111, 200]
//! [22, 122, 200]
//! [33, 133, 200]
//! [44, 144, 200]
//! [55, 155, 200]
//! [66, 166, 200]
//! [77, 177, 200]
//! ...
//! ```

#![no_main]
#![no_std]

extern crate panic_halt;

use rtfm::app;

use driver::pac;
use embedded_hal::blocking::spi::Transfer;
use embedded_hal::spi::{FullDuplex, MODE_0};
use lpc43xx_hal as driver;

type Spi0 = driver::spi::spi::Spi;
type Spi1 = driver::spi::ssp::SpiSsp<pac::SSP1>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut SPI_MASTER: Spi1 = ();
    static mut SPI_SLAVE: Spi0 = ();

    #[init]
    fn init() -> init::LateResources {
        // SSP1 (master) pin configuration
        // SSP1_SCK
        device
            .SCU
            .sfspf_4
            .write(|w| w.mode().function_0_default().epun().enable_pull_up());
        // SSP1_MISO
        device
            .SCU
            .sfspf_6
            .write(|w| w.mode().function_2().ezi().enable_input_buffer());
        // SSP1_MOSI
        device.SCU.sfsp1_4.write(|w| w.mode().function_5());
        // SSP1_SSEL
        device.SCU.sfspf_5.write(|w| w.mode().function_2());

        // SPI (slave) pin configuration
        // SPI_SCK
        device
            .SCU
            .sfsp3_3
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());
        // SPI_MISO
        device.SCU.sfsp3_6.write(|w| w.mode().function_1());
        // SPI_MOSI
        device
            .SCU
            .sfsp3_7
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());
        // SPI_SSEL
        device
            .SCU
            .sfsp3_8
            .write(|w| w.mode().function_1().ezi().enable_input_buffer());

        // slave, enable interrupts
        let spi_slave = Spi0::new(device.SPI, MODE_0, false, true);

        let spi_master = Spi1::new(device.SSP1, MODE_0, true, false);
        init::LateResources {
            SPI_MASTER: spi_master,
            SPI_SLAVE: spi_slave,
        }
    }

    #[idle(resources = [SPI_MASTER])]
    fn idle() -> ! {
        let spi = resources.SPI_MASTER;
        let mut buf = [0u8, 100, 200];
        loop {
            cortex_m::asm::bkpt();
            spi.transfer(&mut buf).unwrap();
            for v in &mut buf {
                *v = v.wrapping_add(1);
            }
            // prepare for the next round
            buf[0] = buf[1];
            buf[1] = buf[2];
            buf[2] = 200;
        }
    }

    #[interrupt(resources = [SPI_SLAVE])]
    fn SPI_INT() {
        let spi = resources.SPI_SLAVE;
        // ideally you would handle potential errors too
        let r = spi.read().unwrap();
        spi.send(r.wrapping_add(10)).unwrap();
    }
};
