//! Timer in counter mode
//! Edges on P1_12 is counted with TIMER0.
//! In a periodic TIMER1 interrupt the counted pulses on p1_12 is read and the
//! counter TIMER0 is reset.
//! Can be used to measure fan speed by counting sensor pulses during fixed time.
#![no_main]
#![no_std]
#[allow(unused_extern_crates)]
extern crate panic_halt;

use embedded_hal::digital::v2::{OutputPin, ToggleableOutputPin};
use lpc43xx_hal::{
    io::pin::{split_ports, Gpio, IntoGpioPin, Output, Pin, PinSplit, U1, U5, U3, U12, Uninit},
    pac, timer,
    timer::{Timer, EMC0_A, MR0I_A, MR0R_A, MR0S_A, NoMatchPin, NoCapturePin},
};
use rtfm::app;
use pac::timer0::ctcr::{CINSEL_A, CTMODE_A};

static BLINK_PERIOD: u32 = 16384;
static TIMER_PRE_SCALE: u32 = 2048;

// used as a counter
type Timer0 = Timer<pac::TIMER0, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, Pin<U1, U12, Uninit>, NoCapturePin, NoCapturePin>;
// used for periodic interrupts
type Timer1 = Timer<pac::TIMER1, NoMatchPin, NoMatchPin, NoMatchPin, NoMatchPin, NoCapturePin, NoCapturePin, NoCapturePin, NoCapturePin>;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut TIMER: Timer1 = ();
    static mut COUNTER: Timer0 = ();
    static mut LED: Pin<U5, U3, Gpio<Output>> = ();
    static mut CNT: u32 = ();

    #[init]
    fn init() -> init::LateResources {
        let ports = split_ports(device.SCU, device.GPIO_PORT);

        // Configure timer
        let mut t = Timer1::new_no_pins(device.TIMER1);
        t.start_timer(false);
        t.set_match(0, BLINK_PERIOD);
        t.set_prescaler(TIMER_PRE_SCALE, Some(0));
        // Safe, because the timer is stopped.
        unsafe {
            t.mr0(MR0I_A::INTERRUPT, MR0R_A::RESET, MR0S_A::NO_STOP, EMC0_A::NOP);
        }
        t.start_timer(true);

        // Configure timer
        let count_pin = ports.port_1.split_pins().p1_12;
        let mut counter = Timer0::new(device.TIMER0, None, None, None, None, None, Some(count_pin), None, None);
        counter.start_timer(false);
        counter.set_ctmode(CTMODE_A::COUNTER_MODE_EDGES, CINSEL_A::CAP1);
        counter.start_timer(true);

        // Map T0_CAP1 to Timer 0 input capture
        device.GIMA.cap0_1_in.write(|w| w
            .inv().not_inverted()
            .edge().rising_edge_detection_enabled()
            .synch().enable_synchroniza()
            .pulse().disable_single_pulse_generation()
            .select().t0_cap1() );

        // Configure GPIO
        let mut led_pin = ports
            .port_5
            .split_pins()
            .p5_3
            .into_gpio()
            .into_output();

        led_pin.set_low().unwrap();
        init::LateResources {
            TIMER: t,
            LED: led_pin,
            COUNTER: counter,
            CNT:0,
        }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [COUNTER, TIMER, CNT, LED])]
    fn TIMER1() {
        resources.LED.toggle().unwrap();
        *resources.CNT = resources.COUNTER.read_tc();
        //Do something with the counted value.
        if *resources.CNT < 0x45 {
            let _cnt = *resources.CNT;
        }
        resources.COUNTER.reset_timer();
        resources.COUNTER.start_timer(true);

        resources
            .TIMER
            .clear_interrupt(timer::InterruptType::MatchChannel, 0);
    }
};
