#![no_main]
#![no_std]
#[allow(unused_extern_crates)]
extern crate panic_halt;

use lpc43xx_hal::{
    io::pin::{split_ports, PinSplit},
    pac,
    serial::{Read, Serial, Write},
};
use pac::usart0::lcr::{BC_A, DLAB_A, PE_A, PS_A, SBS_A, WLS_A};
use rtfm::app;

#[app(device = lpc43xx_pac)]
const APP: () = {
    static mut UART: Serial<pac::USART3> = ();

    #[init]
    fn init() -> init::LateResources {
        // Pins
        let port_9 = split_ports(device.SCU, device.GPIO_PORT)
            .port_9
            .split_pins();
        let uart_tx = port_9.p9_3;
        let uart_rx = port_9.p9_4;

        let mut serial =
            Serial::<pac::USART3>::new(device.USART3, uart_tx, uart_rx, 0x41, 0x0, 0x0);
        serial.lcr(
            WLS_A::_8_BIT_CHARACTER_LENGTH,
            SBS_A::_1_STOP_BIT,
            PE_A::DISABLE_PARITY_GENERERATION,
            PS_A::ODD_PARITY,
            BC_A::DISABLED,
            DLAB_A::DISABLED,
        );
        serial.ier(true, false, false, false, false);

        unsafe {
            // Select the IRC as BASE_M4_CLK source
            device.CGU.base_m4_clk.write(|w| w.clk_sel().bits(0x1));

            // Enable the crystal oscillator
            device.CGU.xtal_osc_ctrl.modify(|_, w| w.hf().bit(false));
            device
                .CGU
                .xtal_osc_ctrl
                .modify(|_, w| w.enable().bit(false));

            // Wait 250us
            for i in 0..3_000 {
                core::ptr::read_volatile(&i);
            }

            device.CGU.pll1_ctrl.write(|w| {
                w.pd()
                    .bit(false)
                    .clk_sel()
                    .bits(0x6) // Select the crystal oscillator as clock source
                    .msel()
                    .bits(20)
                    .nsel()
                    .bits(1) // Select the M and N divider values
                    .psel()
                    .bits(0) // Select the M and N divider values
                    .fbsel()
                    .bit(true)
                    .bypass()
                    .bit(false)
                    .direct()
                    .bit(false)
            });

            // Wait for the PLL to lock
            while !device.CGU.pll1_stat.read().lock().bit() {}

            // Set PLL1 p-divider to divide by 2 (DIRECT=0; PSEL=0)
            device
                .CGU
                .pll1_ctrl
                .modify(|_, w| w.direct().bit(false).psel().bits(0x0));

            // Select PLL1 as base_m4_clk source
            device.CGU.base_m4_clk.modify(|_, w| w.clk_sel().bits(0x09));

            // Wait for 50us
            for i in 0..600 {
                core::ptr::read_volatile(&i);
            }

            // Enable direct output mode
            device.CGU.pll1_ctrl.modify(|_, w| w.direct().bit(true));

            // Enable clock on UART3
            device.CGU.base_uart3_clk.write(|w| w.clk_sel().bits(0x09)); // configure crystal oscillator
        }

        init::LateResources { UART: serial }
    }

    #[idle]
    fn idle() -> ! {
        loop {}
    }

    #[interrupt(resources = [UART])]
    fn USART3() {
        let input = resources.UART.read().unwrap();
        resources.UART.write(input).unwrap();
    }
};
