//! 10-bit (Low-Speed) ADC abstraction
#![deny(unused)]
#![deny(missing_docs)]
use crate::pac;
use pac::SCU;
pub use pac::{ADC0, ADC1};

use crate::io::pin::Pin;
use core::marker::PhantomData;

use embedded_hal::adc::{OneShot, Channel};

#[allow(unused_imports)]
use crate::io::pin::{
    A, B, C, D, E, F, U0, U1, U10, U11, U12, U13, U14, U15, U16, U17, U18, U19, U2, U20, U3, U4,
    U5, U6, U7, U8, U9,
};

use crate::io::pin::{Gpio, Input};

/// ADC pin typestate
pub struct AdcIn <PERIPHERAL>(PhantomData<PERIPHERAL>);

/// Pins that can be muxed from digital to analog inputs
pub trait IntoAdc<X, N, C> {
    /// HAL peripheral
    type Peripheral;

    /// Mux a digital pin into an analog ADC input
    ///
    /// This function is unsafe because it's possible to first allocate the true analog
    /// pin and then mux the digital input onto it as well.
    unsafe fn into_adc(self) -> Pin<X, N, AdcIn<Self::Peripheral>>;
}

/// 10-bit Analog-to-Digital Converter
pub struct Adc<ADC> {
    /// PAC peripheral member
    pub adc: ADC,
}

macro_rules! digital_adc {
    ($($ADC: ident: ($adc: ident, $num: tt): {
        input: [
            $(($tadc_x: tt ($tadc_X: ty), $tadc_n: tt ($tadc_N: ty), $tadc_c: tt ($tadc_C: ty), $adc_sfs_mode: ident),)+
        ],
    })+) => {
        $(
            $(impl IntoAdc<$tadc_X, $tadc_N, $tadc_C> for Pin<$tadc_X, $tadc_N, Gpio<Input>> {
                type Peripheral = $ADC;

                unsafe fn into_adc(self) -> Pin<$tadc_X, $tadc_N, AdcIn<$ADC>> {
                    paste::expr! {
                        // Configure pin state
                        // This is safe because we were given the pin to work with
                        (*SCU::ptr()).[<sfsp $tadc_x _ $tadc_n>].write(|w| w
                            .ezi().clear_bit()
                            .epun().set_bit()
                            .epd().clear_bit()
                            .mode().$adc_sfs_mode());

                        // Set analog input enable
                        // Not safe, because no guarantees are given about the muxed digital pins
                        (*SCU::ptr()).[<enaio $num>].modify(|_, w| w.[<adc$num _ $tadc_c>]().set_bit());
                    };

                    // Safe because we got the pin as an input
                    Pin::<$tadc_X, $tadc_N, AdcIn<$ADC>>::new()
                }
            })+

            impl Adc<$ADC> {
                /// Initialize a new ADC with some clock divider
                pub fn new(device: $ADC, clkdiv: u8) -> Self {
                    let tmp = Self {
                        adc: device,
                    };

                    // Set divider
                    unsafe {
                        tmp.adc.cr.write(|w| w.clkdiv().bits(clkdiv)
                            .pdn().bit(true));
                    }

                    tmp
                }
            }

            $(
                impl Channel<$ADC> for Pin<$tadc_X, $tadc_N, AdcIn<$ADC>> {
                    type ID = (u8, u8);

                    fn channel() -> Self::ID {
                        (0, $tadc_c)
                    }
                }

                impl OneShot<$ADC, u16, Pin<$tadc_X, $tadc_N, AdcIn<$ADC>>> for Adc<$ADC> {
                    type Error = ();

                    fn read(&mut self, _pin: &mut Pin<$tadc_X, $tadc_N, AdcIn<$ADC>>) -> nb::Result<u16, Self::Error> {
                        // Start reading
                        unsafe {
                            self.adc.cr.modify(|_, w| w.sel().bits(1 << $tadc_c));
                            self.adc.cr.modify(|_, w| w.start().bits(0x1));
                        }

                        // Block until the adc is done
                        while self.adc.gdr.read().done().bit() == false {}

                        unsafe {
                            self.adc.cr.modify(|_, w| w.start().bits(0x0));
                        }

                        // Return the channel and the value
                        Ok(self.adc.gdr.read().v_vref().bits() as u16)

                    }
                }
            )+
        )+
    }
}

digital_adc! {
    ADC0: (adc_0, 0): {
        input: [
            (4 (U4), 3  (U3),  0 (U0), function_0_default),
            (4 (U4), 1  (U1),  1 (U1), function_0_default),
            (f  (F), 8  (U8),  2 (U2), function_4),
            (7 (U7), 5  (U5),  3 (U3), function_0_default),
            (7 (U7), 4  (U4),  4 (U4), function_0_default),
            (f  (F), 10 (U10), 5 (U5), function_4),
            (b  (B), 6  (U6),  6 (U6), function_4),
        ],
    }
    ADC1: (adc_1, 1): {
        input: [
            (c  (C), 3   (U3), 0 (U0), function_4),
            // (c  (C), 0   (U0), 1 (U1), ), // No relevant tri-state mode?
            (f  (F), 9   (U9), 2 (U2), function_4),
            (f  (F), 6   (U6), 3 (U3), function_4),
            (f  (F), 5   (U5), 4 (U4), function_4),
            (f  (F), 11 (U11), 5 (U5), function_4),
            (7 (U7), 7   (U7), 6 (U6), function_0_default),
            (f  (F), 7   (U7), 7 (U7), function_4),
        ],
    }
}

macro_rules! analog_adc {
    ($($ADC: ident: ($adc: ident, $num: tt): {
        input: [
            $($NAME: ident, $channel: tt ($CHANNEL: ty),)+
        ],
    })+) => {
        $(
            $(
                /// Pin representation for true analog ADC inputs
                pub struct $NAME;

                impl Channel<$ADC> for Pin<$NAME, $NAME, AdcIn<$ADC>> {
                    type ID = (u8, u8);

                    fn channel() -> Self::ID {
                        (0, $channel)
                    }
                }

                impl OneShot<$ADC, u16, Pin<$NAME, $NAME, AdcIn<$ADC>>> for Adc<$ADC> {
                    type Error = ();

                    /// Simplest possible implementation of ADC read
                    fn read(&mut self, _pin: &mut Pin<$NAME, $NAME, AdcIn<$ADC>>) -> nb::Result<u16, Self::Error> {
                        // Start reading
                        unsafe {
                            self.adc.cr.modify(|_, w| w.sel().bits(1 << $channel));
                            self.adc.cr.modify(|_, w| w.start().bits(0x1));
                        }

                        // Block until the adc is done
                        while self.adc.gdr.read().done().bit() == false {}

                        unsafe {
                            self.adc.cr.modify(|_, w| w.start().bits(0x0));
                        }

                        // Return the channel and the value
                        Ok(self.adc.gdr.read().v_vref().bits() as u16)

                    }
                }
            )+
        )+
    }
}

macro_rules! adc_pin_get {
    ($name: ident, $NAME: ty, $ADC: ty, $CHANNEL: tt) => {
        /// Preview function for using analog ADC pins
        /// NOTE: Here, we check that the pin wasn't previously allocated by a digital muxed pin
        /// Going the other way around would require a more clever abstraction
        /// TODO: To finalize this, we need to move the instantiation its proper place in the io::pin module and
        ///       give it away in the same manner as the rest of the pins
        pub unsafe fn $name(scu: &SCU) -> Pin<$NAME, $NAME, AdcIn<$ADC>> {
            // Ensure that we haven't muxed a digital input from a previous run
            assert!(scu.enaio0.read().bits() & (1 << $CHANNEL) == 0);
            Pin::new()
        }
    }
}

/// Represents a virtual pin "block" from which we can create pins
pub struct ADCBlock;

impl ADCBlock {
    adc_pin_get!(adc0_0, ADC0_0, ADC0, 0);
    adc_pin_get!(adc0_1, ADC0_1, ADC0, 1);
    adc_pin_get!(adc0_2, ADC0_2, ADC0, 2);
    adc_pin_get!(adc0_3, ADC0_3, ADC0, 3);
    adc_pin_get!(adc0_4, ADC0_4, ADC0, 4);
    adc_pin_get!(adc0_5, ADC0_5, ADC0, 5);
    adc_pin_get!(adc0_6, ADC0_6, ADC0, 6);
    adc_pin_get!(adc0_7, ADC0_7, ADC0, 7);
    adc_pin_get!(adc1_0, ADC1_0, ADC1, 0);
    adc_pin_get!(adc1_1, ADC1_1, ADC1, 1);
    adc_pin_get!(adc1_2, ADC1_2, ADC1, 2);
    adc_pin_get!(adc1_3, ADC1_3, ADC1, 3);
    adc_pin_get!(adc1_4, ADC1_4, ADC1, 4);
    adc_pin_get!(adc1_5, ADC1_5, ADC1, 5);
    adc_pin_get!(adc1_6, ADC1_6, ADC1, 6);
    adc_pin_get!(adc1_7, ADC1_7, ADC1, 7);
}

analog_adc! {
    ADC0: (adc_0, 0): {
        input: [
            ADC0_0, 0 (U0),
            ADC0_1, 1 (U1),
            ADC0_2, 2 (U2),
            ADC0_3, 3 (U3),
            ADC0_4, 4 (U4),
            ADC0_5, 5 (U5),
            ADC0_6, 6 (U6),
            ADC0_7, 7 (U7),
        ],
    }
    ADC1: (adc_1, 0): {
        input: [
            ADC1_0, 0 (U0),
            ADC1_1, 1 (U1),
            ADC1_2, 2 (U2),
            ADC1_3, 3 (U3),
            ADC1_4, 4 (U4),
            ADC1_5, 5 (U5),
            ADC1_6, 6 (U6),
            ADC1_7, 7 (U7),
        ],
    }
}
