//! An interface to the two CAN-bus controllers present on the device.
//! According to UM10503 the controllers conform to protocol version 2A/B
use core::cmp;
use heapless::{Vec, ArrayLength};

use super::MsgObj;
use super::{C_CAN0, C_CAN1};
use super::{IntoCanRx, IntoCanTx, CanRx, CanTx};

use super::{CCE_A, DAR_A, EIE_A, IE_A, SIE_A, TEST_A};
use super::LEC_A;
use super::{BASIC_A, LBACK_A, RX_A, SILENT_A};

/// One CAN-bus peripheral
pub struct CanBus<CAN, N: ArrayLength<u8>> {
    pub test_configured: bool,
    pub can: CAN,
    pub data_buf: Vec<u8, N>,
}

/// Message RAM interfaces
pub enum MsgIface {
    Iface1,
    Iface2,
}

/// Message RAM 
pub enum MsgDir {
    Rx = 0b0,
    Tx = 0b1,
}

macro_rules! hal {
    ($($CAN: ident: ($can: ident),)+) => {
        $(
            impl<N: ArrayLength<u8>> CanBus<$CAN, N> {
                pub fn new<XTx, NTx, Tx, XRx, NRx, Rx>(device: $CAN, tx_pin: Tx, rx_pin: Rx) -> Self
                    where
                        Tx: IntoCanTx<XTx, NTx, Peripheral=$CAN>,
                        Rx: IntoCanRx<XRx, NRx, Peripheral=$CAN> {
                    // Pin config
                    tx_pin.into_can_tx();
                    rx_pin.into_can_rx();

                    Self {
                        can: device,
                        test_configured: false,
                        data_buf: heapless::Vec::new(),
                    }
                }

                /// Determine if the CAN controller was initialized
                fn initializing(&self) -> bool {
                    self.can.cntl.read().init().bit()
                }

                /// Retrieve interrupt ID
                ///
                /// Reads the current interrupt ID for use in interrupt handler
                pub fn read_intid(&self) -> u16 {
                    (self.can.int.read().intid15_0().bits() & 0xffff) as u16
                }

                /// Get passive error level
                ///
                /// The receive counter has reached the error passive level as
                /// defined in the CAN 2.0 specification
                pub fn get_error_level(&self) -> bool {
                    self.can.ec.read().rp().bit()
                }

                /// Read the Transmit Error Counter
                pub fn get_tec(&self) -> u8 {
                    self.can.ec.read().tec_7_0().bits() & 0xff
                }

                /// Read the Receive Error Counter
                pub fn get_rec(&self) -> u8 {
                    self.can.ec.read().rec_6_0().bits() & 0x7f
                }

                // /// Get last error code
                // pub fn get_lec(&self) -> FR {
                //     self.can.stat.read().lec()
                // }

                /// Set error code to 'unused'
                pub fn clear_lec(&self) {
                    self.can.stat.write(|w| w.lec().bits(0x7));
                }

                /// Block until TX complete
                pub fn tx_block(&self, idx: u8) {
                    loop {
                        let bits: u32 = self.can.txreq1.read().bits() | self.can.txreq2.read().bits() << 16;
                        if bits & (1 << (idx - 1)) == 0 {
                            return;
                        }
                    }
                }

                /// Set the clock divider
                pub fn clk_div(&mut self, value: u8) {
                    assert!(value & 0xf0 == 0);
                    unsafe {
                        self.can.clkdiv.write(|w| w.clkdivval().bits(value));
                    }
                }

                /// Find bitrate parameters
                #[allow(unused_assignments)]
                fn bitrate_params(&mut self, input_rate: u32, bitrate: u32) -> Result<(u32, u32, u8, u8, u8), &'static str> {
                    let clk_per_bit = input_rate/bitrate;
                    let mut can_sjw: u32 = 0x0;
                    let mut seg1: u32 = 0x0;
                    let mut seg2: u32 = 0x0;

                    for div in 0..=15 {
                        for quanta in 1..=32 {
                            for mut segs in 3..=17 {
                                if clk_per_bit == (segs * quanta * (div + 1)) {
                                    segs -= 3;
                                    seg1 = segs / 2;
                                    seg2 = segs - seg1;
                                    can_sjw = cmp::min(seg1, 3);
                                    return Ok((div, quanta - 1, can_sjw as u8, seg1 as u8, seg2 as u8))
                                }
                            }
                        }
                    }

                    Err("failed to find bitrate")
                }

                /// Adjust the sample point and synchronization jump width
                fn config_timing(&self, brp: u8, sjw: u8, tseg1: u8, tseg2: u8) {
                    assert!(brp & 0xc0 == 0);
                    assert!(sjw < 4);
                    assert!(tseg1 < 8);
                    assert!(tseg2 > 0 && tseg2 < 16);
                    self.can.cntl.modify(|_, w| w.cce().bit(true));
                    unsafe {
                        self.can.bt.modify(|_, w|
                                 w.brp().bits(brp)
                                 .sjw().bits(sjw)
                                 .tseg1().bits(tseg1)
                                 .tseg2().bits(tseg2));
                    }
                    self.can.cntl.modify(|_, w| w.cce().bit(false));
                }

                /// Configure test parameters
                pub fn config_test(&mut self, basic: BASIC_A, silent: SILENT_A, loopback: LBACK_A, tx_ctrl: u8, rx_ctrl: RX_A) {
                    assert!(tx_ctrl < 4);
                    self.test_configured = true;

                    self.can.test.write(|w| w
                        .basic()
                        .variant(basic)
                        .silent()
                        .variant(silent)
                        .lback()
                        .variant(loopback)
                        .tx1_0()
                        .bits(tx_ctrl)
                        .rx()
                        .variant(rx_ctrl));
                }

                /// Initialize the CAN controller
                pub fn initialize(
                &mut self,
                input_frequency: u32,
                bitrate: u32,
                modinten:     IE_A,
                scinten:      SIE_A,
                errinten:     EIE_A,
                disable_retr: DAR_A,
                test:         bool) {
                    self.can.cntl.write(|w| w
                        .init().bit(true)
                        .cce()
                        .variant(CCE_A::THE_CPU_HAS_WRITE_ACCESS));

                    // Setup bit timing register
                    let (div, quanta, sjw, seg1, seg2) =
                        self.bitrate_params(input_frequency, bitrate).unwrap();
                    self.config_timing(quanta as u8, sjw, seg1, seg2);
                    self.clk_div(div as u8);

                    for i in 1..=0x20 {
                        self.clear_ram(MsgIface::Iface1, i);
                        self.clear_ram(MsgIface::Iface2, i);
                    }

                    // Finish initialization
                    self.can.cntl.write(|w| w.ie().variant(modinten)
                    .sie()
                    .variant(scinten)
                    .eie()
                    .variant(errinten)
                    .dar()
                    .variant(disable_retr)
                    .cce()
                    .variant(CCE_A::THE_CPU_HAS_NO_WRITE_ACCESS)
                    .test()
                    .variant(if test {
                            if self.test_configured {
                                TEST_A::TEST_MODE
                            } else {
                                panic!("error: test not configured, but test initialized!")
                            }
                        } else { TEST_A::NORMAL_OPERATION }
                    )
                    .init().bit(false));
                }

                /// Read and clear status interrupt field
                pub fn read_status(&mut self) -> crate::pac::c_can1::stat::R {
                    self.can.stat.read()
                }

                /// Receive one or more object and place it in owned memory
                ///
                /// This procedure was implemented from the flowchart in figure 156 of UM10503.
                /// Data is output by filling the data_buf vector
                #[inline(always)]
                pub fn receive_obj(&mut self, iface: MsgIface) -> Result<Vec<u8, N>, Result<u32, LEC_A>> {
                    let intid: u16 = self.read_intid(); //(self.can.int.read().intid15_0().bits() & 0xffff) as u16;
                    self.data_buf.clear(); // Clear output

                    loop {
                        if intid == 0x0 { // INTID = 0x0000 ?
                            return Ok(self.data_buf.clone()); // END
                        } else if intid == 0x8000 { // INTID = 0x8000 ?
                            // Clear status interrupt
                            let stat = self.read_status();

                            if stat.boff().bit() {
                                panic!("BUSOFF");
                            }
                            return Err(Ok(
                                self.can.stat.read().bits() |
                                self.can.ec.read().bits() << 16
                                ))
                        } else if intid > 0x20 {
                            panic!("invalid interrupt ID: {}", intid);
                        } else { // INTID = 0x0001 to 0x0020 ?
                            let mut message_num: u8 = intid as u8; // MessageNum = INTID

                            loop {
                                macro_rules! receive_data {
                                    ($cmdmsk: ident, $cmdmsk_r: ident, $cmdreq: ident, $mctrl: ident, $da1: ident,
                                        $da2: ident, $db1: ident, $db2: ident) => {
                                        unsafe {
                                            // read message to message buffer
                                            // P. 1094 instructions on how to read message into buffer
                                            self.can.$cmdmsk().write(|w| w.bits(0x7f));

                                            // write MessageNum to CANIFx_CMDREQ
                                            self.can.$cmdreq.write(|w| w.messnum().bits(message_num));

                                            // NOTE: Clear RXOK
                                            self.can.stat.modify(|_, w| w.rxok().bit(false));

                                            (
                                                self.can.$mctrl.read().newdat().bit(),
                                                self.can.$mctrl.read().eob().bit(),
                                                self.can.$mctrl.read().dlc3_0().bits(),
                                                ((self.can.$da1.read().bits() as u32) & 0xffff) |
                                                    ((self.can.$da2.read().bits() as u32) << 16),
                                                ((self.can.$db1.read().bits() as u32) & 0xffff) |
                                                    ((self.can.$db2.read().bits() as u32) << 16),
                                            )

                                        };
                                    }
                                }

                                // Gather all messages
                                let (newdat, eob, dlc, data_1, data_2) = match iface {
                                    MsgIface::Iface1 =>
                                        receive_data!(if1_cmdmsk_r, if1_cmdmsk_r, if1_cmdreq, if1_mctrl, if1_da1,
                                            if1_da2, if1_db1, if1_db2),
                                    MsgIface::Iface2 =>
                                        receive_data!(if2_cmdmsk_r, if2_cmdmsk_r, if2_cmdreq, if2_mctrl, if2_da2,
                                            if2_da2, if2_db2, if2_db2),
                                };

                                macro_rules! clear_bits {
                                    ($cmdmsk: ident, $cmdmsk_r: ident, $cmdreq: ident) => {
                                        // reset NEWDAT = 0
                                        // reset INTPND = 0
                                        unsafe {
                                            self.can.$cmdmsk_r().write(|w| w.newdat().bit(true)
                                                .clrintpnd().bit(true)
                                                .wr_rd().bit(true));

                                            // write MessageNum to CANIFx_CMDREQ
                                            self.can.$cmdreq.write(|w| w.messnum().bits(message_num));
                                            while self.can.$cmdreq.read().busy().bit() {}
                                        }
                                    }
                                }

                                // Reset newdata and intpnd
                                match iface {
                                    MsgIface::Iface1 => clear_bits!(if1_cmdmsk, if1_cmdmsk_r, if1_cmdreq),
                                    MsgIface::Iface2 => clear_bits!(if2_cmdmsk, if2_cmdmsk_r, if2_cmdreq),
                                };

                                if !newdat { // No data in the message
                                    return Ok(self.data_buf.clone())
                                } else {
                                    // Copy received data
                                    let true_dlc = if dlc > 8 { 8 } else { dlc };
                                    for i in 0..true_dlc {
                                        match i {
                                            0 ..= 3 => {
                                                let to_push = data_1 >> (i << 3);
                                                self.data_buf.push(to_push as u8).unwrap();
                                            },
                                            4 ..= 7 => {
                                                let to_push = data_2 >> ((i-4) << 3);
                                                self.data_buf.push(to_push as u8).unwrap();
                                            },
                                            _ => panic!("invalid dlc {}", i),
                                        };
                                    }

                                    if eob { // EOB = 1
                                        return Ok(self.data_buf.clone())
                                    } else {
                                        // NOTE: this part is untested
                                        message_num += 1; // MessageNum = MessageNum +1
                                        continue;
                                    }
                                }
                            }
                        }
                    }
                }

                /// Clear message RAM on specified index
                pub fn clear_ram(&mut self, iface: MsgIface, idx: u8) {
                    assert!(idx >= 0x01 && idx <= 0x20);
                    let mut obj = MsgObj::default();
                    obj.cmdmsk = 0xff & !0x4;
                    Self::transfer_obj(self, iface, idx, &obj);
                }

                /// Transfers an object
                ///
                /// Copies an object from the program RAM to the CAN-bus message RAM
                pub fn transfer_obj(&mut self, iface: MsgIface, number: u8, obj: &MsgObj) {
                    assert!(number >= 0x01 && number <= 0x20);

                    macro_rules! transfer {
                        ($cmdmsk: ident, $cmdmsk_w: ident, $msk1: ident, $msk2: ident, $arb1: ident, $arb2: ident,
                            $mctrl: ident, $da1: ident, $da2: ident, $db1: ident, $db2: ident, $cmdreq: ident) => {
                            unsafe {
                                self.can.$cmdmsk_w().write(|w| w.bits(obj.cmdmsk));
                                self.can.$msk1.write(|w| w.bits(obj.mask_1));
                                self.can.$msk2.write(|w| w.bits(obj.mask_2));

                                // Set arbitration parameters
                                self.can.$arb1.write(|w| w.bits(obj.arb_1));
                                self.can.$arb2.write(|w| w.bits(obj.arb_2));

                                self.can.$mctrl.write(|w| w.bits(obj.mctrl));

                                // Write data
                                self.can.$da1.write(|w| w.bits(obj.data_a1 as u32));
                                self.can.$da2.write(|w| w.bits(obj.data_a2 as u32));
                                self.can.$db1.write(|w| w.bits(obj.data_b1 as u32));
                                self.can.$db2.write(|w| w.bits(obj.data_b2 as u32));

                                // Request transfer and wait for complete
                                self.can.$cmdreq.write(|w| w.bits(number as u32));
                                while self.can.$cmdreq.read().busy().bit() {}
                            }
                        }
                    }

                    match iface {
                        MsgIface::Iface1 => transfer!(if1_cmdmsk, if1_cmdmsk_w, if1_msk1, if1_msk2, if1_arb1,
                            if1_arb2, if1_mctrl, if1_da1, if1_da2, if1_db1, if1_db2, if1_cmdreq),
                        MsgIface::Iface2 => transfer!(if2_cmdmsk, if2_cmdmsk_w, if2_msk1, if2_msk2, if2_arb1,
                            if2_arb2, if2_mctrl, if2_da1, if2_da2, if2_db1, if2_db2, if2_cmdreq),
                    };
                }
            }
        )+
    }
}

hal! {
    C_CAN0: (c_can0),
    C_CAN1: (c_can1),
}
