pub use crate::pac::c_can1::if_cmdmsk_w::*;

#[derive(Debug)]
pub struct MsgObj {
    pub cmdmsk: u32,
    pub mctrl: u32,
    pub mask_1: u32,
    pub mask_2: u32,
    pub data_a1: u16,
    pub data_a2: u16,
    pub data_b1: u16,
    pub data_b2: u16,
    pub arb_1: u32,
    pub arb_2: u32,
}

impl Default for MsgObj {
    fn default() -> Self {
        MsgObj {
            cmdmsk: 0x0,
            mctrl: 0x0,
            mask_1: 0x0,
            mask_2: 0x0,
            data_a1: 0x0,
            data_a2: 0x0,
            data_b1: 0x0,
            data_b2: 0x0,
            arb_1: 0x0,
            arb_2: 0x0,
        }
    }
}

impl MsgObj {
    /// Create a basic transmit object
    pub fn new_tx_obj(txie: bool, id: u16, length: u8, data_a: u32, data_b: u32) -> MsgObj {
        assert!(length <= 0x8);
        let mut tx_obj = MsgObj::default();
        tx_obj
            .dlc(length)
            .eob(true)
            .dir(true)
            .newdat(true)
            .msglst(false)
            .txie(txie)
            .rxie(false)
            .intpnd(false)
            .id_low(0x0)
            .id_high(id << 2)
            .data_b(false)
            .data_a(false)
            .cmdmsk_txrqst(false)
            .ctrl(true)
            .arb(true)
            .mask(true)
            .wr_rd(true);

        tx_obj.data_a1(data_a as u8, (data_a >> 8) as u8);
        tx_obj.data_a2((data_a >> 16) as u8, (data_a >> 24) as u8);
        tx_obj.data_b1(data_b as u8, (data_b >> 8) as u8);
        tx_obj.data_b2((data_b >> 16) as u8, (data_b >> 24) as u8);

        tx_obj
    }

    /// Create a basic receive object
    pub fn new_rx_obj(rxie: bool, id: u16, length: u8, mask: u32) -> MsgObj {
        let mut rx_obj = MsgObj::default();
        rx_obj
            .dlc(length)
            .msgval(true)
            .eob(true)
            .dir(false)
            .newdat(false)
            .msglst(false)
            .txie(false)
            .rxie(rxie)
            .intpnd(false)
            .rmten(false)
            .mctrl_txrqst(false)
            .id_low(0x0)
            .id_high(id << 2)
            .data_b(false)
            .data_a(false)
            .cmdmsk_txrqst(false)
            .ctrl(true)
            .arb(true)
            .mask(true)
            .wr_rd(true)
            .mask_1_id(0xffff)
            .mask_2_id((mask & 0x1fff) as u16);

        rx_obj
    }
}

macro_rules! set_int {
    ($prop: expr, $mask: expr, $value: expr) => {
        $prop = ($prop & !($mask)) | ($value & $mask);
    };
}

macro_rules! set_flag {
    ($prop: expr, $mask: expr, $value: expr) => {
        $prop = ($prop & !($mask)) | if $value { ($mask) } else { 0 };
    };
}

// CMDMSK fields
impl MsgObj {
    pub fn data_b<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 0, value);
        self
    }
    pub fn data_a<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 1, value);
        self
    }
    pub fn cmdmsk_txrqst<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 2, value);
        self
    }
    pub fn clrintpnd<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 3, value);
        self
    }
    pub fn ctrl<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 4, value);
        self
    }
    pub fn arb<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 5, value);
        self
    }
    pub fn mask<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 6, value);
        self
    }
    pub fn wr_rd<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.cmdmsk, 1 << 7, value);
        self
    }
}
// MCTRL fields
impl MsgObj {
    pub fn dlc<'a>(&'a mut self, value: u8) -> &mut Self {
        assert!(value & 0xf0 == 0);
        set_int!(self.mctrl, 0x0f, value as u32);
        self
    }

    pub fn eob<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 7, value);
        self
    }

    pub fn mctrl_txrqst<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 8, value);
        self
    }
    pub fn rmten<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 9, value);
        self
    }
    pub fn rxie<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 10, value);
        self
    }
    pub fn txie<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 11, value);
        self
    }
    pub fn umask<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 12, value);
        self
    }
    pub fn intpnd<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 13, value);
        self
    }

    pub fn msglst<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 14, value);
        self
    }
    pub fn newdat<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mctrl, 1 << 15, value);
        self
    }
}

// Mask 1 fields
impl MsgObj {
    pub fn mask_1_id<'a>(&'a mut self, value: u16) -> &mut Self {
        set_int!(self.mask_1, 0xffff, value as u32);
        self
    }
}

// Mask 2 fields
impl MsgObj {
    pub fn mask_2_id<'a>(&'a mut self, value: u16) -> &mut Self {
        set_int!(self.mask_2, 0x1fff, value as u32);
        self
    }
    pub fn mdir<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mask_2, 1 << 14, value);
        self
    }
    pub fn mxtd<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.mask_2, 1 << 15, value);
        self
    }
}

// Data
impl MsgObj {
    pub fn data_a1<'a>(&'a mut self, a: u8, b: u8) -> &mut Self {
        self.data_a1 = ((b as u16) << 8) | (a as u16);
        self
    }
    pub fn data_a2<'a>(&'a mut self, a: u8, b: u8) -> &mut Self {
        self.data_a2 = ((b as u16) << 8) | (a as u16);
        self
    }
    pub fn data_b1<'a>(&'a mut self, a: u8, b: u8) -> &mut Self {
        self.data_b1 = ((b as u16) << 8) | (a as u16);
        self
    }
    pub fn data_b2<'a>(&'a mut self, a: u8, b: u8) -> &mut Self {
        self.data_b2 = ((b as u16) << 8) | (a as u16);
        self
    }
}

// Arbitration 1 fields
impl MsgObj {
    pub fn id_low<'a>(&'a mut self, value: u16) -> &mut Self {
        set_int!(self.arb_1, 0xffff, value as u32);
        self
    }
}

// Arbitration 2 fields
impl MsgObj {
    pub fn id_high<'a>(&'a mut self, value: u16) -> &mut Self {
        set_int!(self.arb_2, 0x1fff, value as u32);
        // self.arb_2 = (self.arb_2 & (!0xfff)) | (value & 0xfff);
        self
    }
    pub fn dir<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.arb_2, 1 << 13, value);
        self
        // self.arb_2 = (self.arb_2 & !(1 << 13)) | if value { (1 << 13) } else { 0 };
    }
    pub fn xtd<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.arb_2, 1 << 14, value);
        self
    }

    pub fn msgval<'a>(&'a mut self, value: bool) -> &mut Self {
        set_flag!(self.arb_2, 1 << 15, value);
        self
    }
}
