//! Pin abstraction for CanBus
use super::{SCU, C_CAN0, C_CAN1};
use crate::io::pin::{Pin, U0, U1, U2, U3, U4, U8, U9, U17, U18, E};
use core::marker::PhantomData;


/// Type states
pub struct CanTx<PERIPHERAL>(PhantomData<PERIPHERAL>);
pub struct CanRx<PERIPHERAL>(PhantomData<PERIPHERAL>);

/// Pins that can be used for communications over CAN
pub trait IntoCanTx<X, N> {
    /// Our HAL peripheral
    type Peripheral;

    /// Create a canbus pin
    fn into_can_tx(self) -> Pin<X, N, CanTx<Self::Peripheral>>;
}

/// Pins that can be used for communications over CAN
pub trait IntoCanRx<X, N> {
    /// Our HAL peripheral
    type Peripheral;

    /// Create a canbus pin
    fn into_can_rx(self) -> Pin<X, N, CanRx<Self::Peripheral>>;
}

macro_rules! can_input_pins {
    ($($CAN: ident: $can: ident {
        read: [
            $(($tcan_x: tt ($tcan_X: ty), $tcan_n: tt ($tcan_N: ty), $sfs_mode: ident),)+
        ],
    })+) => {
        $(
            $(impl<MODE> IntoCanRx<$tcan_X, $tcan_N> for Pin<$tcan_X, $tcan_N, MODE> {
                type Peripheral = $CAN;

                fn into_can_rx(self) -> Pin<$tcan_X, $tcan_N, CanRx<$CAN>> {
                    // Safety: Modifying sfspx_n is safe because the register is in the SCU, which is covered by `Pin`
                    //   and can only be used to configure pin (x, n).
                    unsafe {
                        paste::expr! {
                            (*SCU::ptr()).[<sfsp $tcan_x _ $tcan_n>].modify(|_, w| {
                                w.epun().disable_pull_up()
                                .ezi().set_bit()
                                .mode().$sfs_mode()
                            });
                        }

                        Pin::<$tcan_X, $tcan_N, CanRx<$CAN>>::new()
                    }
                }
            })+
        )+
    }
}

macro_rules! can_output_pins {
    ($($CAN: ident: $can: ident {
        transmit: [
            $(($tcan_x: tt ($tcan_X: ty), $tcan_n: tt ($tcan_N: ty), $sfs_mode: ident),)+
        ],
    })+) => {
        $(
            $(impl<MODE> IntoCanTx<$tcan_X, $tcan_N> for Pin<$tcan_X, $tcan_N, MODE> {
                type Peripheral = $CAN;

                fn into_can_tx(self) -> Pin<$tcan_X, $tcan_N, CanTx<$CAN>> {
                    // Safety: Modifying sfspx_n is safe because the register is in the SCU, which is covered by `Pin`
                    //   and can only be used to configure pin (x, n).
                    unsafe {
                        paste::expr! {
                            (*SCU::ptr()).[<sfsp $tcan_x _ $tcan_n>].modify(|_, w| {
                                w.epun().disable_pull_up()
                                .mode().$sfs_mode()
                            });
                        }

                        Pin::<$tcan_X, $tcan_N, CanTx<$CAN>>::new()
                    }
                }
            })+
        )+
    }
}

can_input_pins! {
    C_CAN0: c_can0 {
        read: [
            (3 (U3), 1 (U1), function_2),
            (e ( E), 2 (U2), function_1),
        ],
    }
    C_CAN1: c_can1 {
        read: [
            (1 (U1), 18 (U18), function_5),
            (4 (U4), 9 (U9), function_6),
            (e ( E), 1 (U1), function_5),
        ],
    }
}

can_output_pins! {
    C_CAN0: c_can0 {
        transmit: [
            (3 (U3), 2 (U2), function_2),
            (e ( E), 3 (U3), function_1),
        ],
    }
    C_CAN1: c_can1 {
        transmit: [
            (1 (U1), 17 (U17), function_5),
            (4 (U4), 8 (U8), function_6),
            (e ( E), 0 (U0), function_5),
        ],
    }
}


