//! I2C (I-squared-C) abstraction for I2C-capable peripherals
//! Only Master mode is currently supported
use embedded_hal::blocking::i2c as hal_i2c;

use crate::pac;
use crate::time::{MegaHertz, KiloHertz, Hertz};

// I2C Speeds
pub const SPEED_NORMAL: KiloHertz = KiloHertz(100);
pub const SPEED_FAST: KiloHertz = KiloHertz(400);
pub const SPEED_FAST_PLUS: MegaHertz = MegaHertz(1);

// I2C status codes
pub const STAT_MSTR_STA: u8 = 0x08 >> 3;
pub const STAT_MSTR_REPSTA: u8 = 0x10 >> 3;
pub const STAT_MSTR_ARB_LOSS: u8 = 0x38 >> 3;
pub const STAT_MSTR_BUS_ERR: u8 = 0x00;

// Status codes which can be returned by the 
// i2c peripheral in master transmitter mode
pub const STAT_MSTR_TRS_SLAW_ACK: u8 = 0x18 >> 3;
pub const STAT_MSTR_TRS_SLAW_NACK: u8 = 0x20 >> 3;
pub const STAT_MSTR_TRS_DAT_ACK: u8 = 0x28 >> 3;
pub const STAT_MSTR_TRS_DAT_NACK: u8 = 0x30 >> 3;

// Status codes which can be returned by the 
// i2c peripheral in master receive mode
pub const STAT_MSTR_RCV_SLAR_ACK: u8 = 0x40 >> 3;
pub const STAT_MSTR_RCV_SLAR_NACK: u8 = 0x48 >> 3;
pub const STAT_MSTR_RCV_DAT_ACK: u8 = 0x50 >> 3;
pub const STAT_MSTR_RCV_DAT_NACK: u8 = 0x58 >> 3;

/// Some standardized speeds which I2C is normally run on.
#[derive(PartialEq)]
pub enum Speed {
    Normal,
    Fast,
    FastPlus,
    Specific(KiloHertz),
}

/// Different I2C Modes.
#[derive(PartialEq, Debug)]
pub enum Mode {
    Master,
    Slave,
}

/// Errors the abstraction can return.
#[derive(PartialEq, Debug)]
pub enum Error {
    UnexpectedNack,
    UnexpectedAck,
    Bus,
    UnknownState(u8)
}

/// I2C Abstraction which is implementing embedded_hal's
/// I2C blocking Read, and Write traits
#[allow(dead_code)]
pub struct I2C {
    peripheral: pac::I2C0,
    speed: Speed,
}

impl I2C {
    
    /// Configure the I2C peripheral.
    /// Only Master mode is supported.
    pub fn new(
        pins: &pac::scu::SFSI2C0,
        peripheral: pac::I2C0, 
        mode: Mode,
        speed: Speed,
        pclk: MegaHertz,
    ) -> Self {

        assert!(
            mode == Mode::Master,
            "Master mode is only supported!");
        
        let bitrate = match speed {
            Speed::Normal => SPEED_NORMAL,
            Speed::Fast => SPEED_FAST,
            Speed::FastPlus => SPEED_FAST_PLUS.into(),
            Speed::Specific(khz) => { khz }
        };

        let br_hz: Hertz = bitrate.into();
        let pclk_hz: Hertz = pclk.into();
        let total_hold = (pclk_hz.0 / br_hz.0)  as u16;

        let use_fp_mode = br_hz.0 > u32::from(SPEED_FAST);

        // Configure pins
        pins.modify(|_, w| { 
            // Use "fastmode plus" if speed > 400KHz
            if use_fp_mode {
                w.
                    scl_ehd().fast_mode_plus_transmit().
                    sda_ehd().fast_mode_plus_transmit();
            } else {
                w.
                    scl_ehd().standardfast_mode().
                    sda_ehd().standardfast_mode();
            }

            // enable input receiver on SDA,SDL pin
            w.
                scl_ezi().enabled().
                sda_ezi().enabled()
        });

        // Initialize i2c
        peripheral.conset.write(|w| {
            w.
                // enable i2c peripheral
                i2en().set_bit()
        });

        // Assume 50% duty cycle on bitrate
        peripheral.sclh.modify(|_, w| unsafe {
            w.
                // Set high hold time to half of bitrate
                sclh().bits(total_hold / 2)
        });
        peripheral.scll.modify(|_, w| unsafe {
            w.
                // Set low hold time to half of bitrate
                scll().bits(total_hold / 2)
        });

        // Prepare master mode
        peripheral.conclr.write(|w| {
            w.
                // clear interrupt bit
                sic().set_bit().
                // clear ack bit
                aac().set_bit().
                // clear start bit
                stac().set_bit()
        });
            
        I2C {
            peripheral: peripheral,
            speed: speed,
        }
    }

    /// Release the peripheral.
    pub fn release(self) -> pac::I2C0 {
        self.peripheral
    }

    fn wait_for_interrupt(&self) {
        while !self.peripheral.conset.read().si().bit_is_set() {}
    }
    fn clear_interrupt(&self) {
        self.peripheral.conclr.write(|w| w.sic().set_bit());
    }
    fn clear_start(&self) {
        self.peripheral.conclr.write(|w| w.stac().set_bit());
    }
    fn read_status(&self) -> u8 {
        self.peripheral.stat.read().status().bits()
    }

    fn begin(&self, addr: u8) -> Result<(), Error> {

        // Send start bit
        self.peripheral.conset.write(|w| {
            w.
                sta().set_bit()
        });

        self.wait_for_interrupt();

        let status = self.read_status();
        match status {
            // (Repeated) Start bit has been sent
            STAT_MSTR_STA | STAT_MSTR_REPSTA => { Ok(()) },
            STAT_MSTR_BUS_ERR => { Err(Error::Bus) },
            _ => { Err(Error::UnknownState(status)) },
        }?;

        // Write address to be sent
        self.peripheral.dat.modify(|_, w| unsafe {
            w.
                data().bits(addr)
        });

        // Clear start bit and clear
        // interrupt to continue
        self.clear_start();
        self.clear_interrupt();

        Ok(())
    }
}

impl hal_i2c::Write for I2C {

    type Error = Error;

    fn write(&mut self, addr: u8, bytes: &[u8]) -> Result<(), Self::Error> {

        // LSB should be 0 for writing
        self.begin((addr << 1) + 0)?;

        let data_len = bytes.len();

        self.wait_for_interrupt();

        let status = self.read_status();
        match status {
            
            STAT_MSTR_TRS_SLAW_ACK => {
                if data_len == 0 {
                    Err(Error::UnexpectedAck)
                } else { Ok(()) }
            },
            STAT_MSTR_TRS_SLAW_NACK => {
                if data_len > 0 {
                    Err(Error::UnexpectedNack)
                } else { Ok(()) }
            },
            STAT_MSTR_BUS_ERR => { Err(Error::Bus) },
            _ => { Err(Error::UnknownState(status)) },

        }?;

        // Send data bytes
        for (i, data) in bytes.iter().enumerate() {

            // Load data byte
            self.peripheral.dat.modify(|_, w| unsafe {
                w.
                    data().bits(*data)
            });

            // Clear interrupt and continue
            self.clear_interrupt();

            self.wait_for_interrupt();

            let status = self.read_status();
            match status {
                
                 STAT_MSTR_TRS_DAT_ACK => { Ok(()) },
                 STAT_MSTR_TRS_DAT_NACK => {
                     // If nack is recieved while more bytes needs
                     // to be written on the bus
                     if i < (data_len - 1) {
                        Err(Error::UnexpectedAck)
                     } else { Ok(()) }
                 },
                STAT_MSTR_BUS_ERR => { Err(Error::Bus) },
                _ => { Err(Error::UnknownState(status)) },
            }?;
        }

        // Send stop bit after all bytes are sent
        self.peripheral.conset.write(|w| {
            w.
                sto().set_bit()
        });

        // And continue
        self.clear_interrupt();

        Ok(())
    }
}

impl hal_i2c::Read for I2C {

    type Error = Error;

    fn read(&mut self, addr: u8, buffer: &mut [u8]) -> Result<(), Self::Error> {

        // LSB should be 1 for writing
        self.begin((addr << 1) + 1)?;

        let buffer_len = buffer.len();

        self.wait_for_interrupt();

        let status = self.read_status();
        match status {
            STAT_MSTR_RCV_SLAR_ACK => { Ok(()) }
            STAT_MSTR_RCV_SLAR_NACK => {
                Err(Error::UnexpectedNack)
            },
            STAT_MSTR_BUS_ERR => { Err(Error::Bus) },
            _ => { Err(Error::UnknownState(status)) },
        }?;

        // Loop through buffer and read into it
        for (i, d) in buffer.iter_mut().enumerate() { 
            let is_last = i == (buffer_len - 1);

            // set acknowledgement bit if it is
            // not the last byte to be received
            if !is_last {
                self.peripheral.conset.modify(|_, w| {
                    w.
                        aa().set_bit()
                });
            } else {
                    self.peripheral.conclr.write(| w| {
                    w.
                        aac().set_bit()
                });
            }

            // clear interrupt and continue
            self.clear_interrupt();

            // wait for interrupt
            self.wait_for_interrupt();

            // this needs to be FIXED
            // wait for data to be recieved
            let status = self.read_status();
            match status {
                STAT_MSTR_RCV_DAT_ACK => {
                    if is_last {
                        Err(Error::UnexpectedAck)
                    } else { Ok(()) }
                },
                STAT_MSTR_RCV_DAT_NACK => {
                    if !is_last {
                        Err(Error::UnexpectedNack)
                    } else { Ok(()) }
                },
                STAT_MSTR_BUS_ERR => { Err(Error::Bus) },
                _ => { Err(Error::UnknownState(status)) },
            }?;

            // Read data and put it into the buffer
            *d = self.peripheral.dat.read().data().bits();
        }

        // send stop
        self.peripheral.conset.modify(|_, w| {
            w.
                sto().set_bit()
        });

        // clear interrupt flag
        self.clear_interrupt();

        Ok(())
    }
}
