use crate::pac;
pub use pac::sct::config::{CKSEL_A, CLKMODE_A, UNIFY_A};

use super::util::CtrlBlock;
use crate::unified;

/// Setup the SCT clock mode, register unification and behaviour
pub fn setup(
    sct: &pac::SCT,
    unify: UNIFY_A,
    clk_mode: CLKMODE_A,
    clk_sel: CKSEL_A,
    no_reload_lower: bool,
    no_reload_higher: bool,
    input_sync: u8,
) -> () {
    unsafe {
        sct.config.write(|w| {
            w.unify()
                .variant(unify)
                .clkmode()
                .variant(clk_mode)
                .cksel()
                .variant(clk_sel)
                .norelaod_l()
                .bit(no_reload_lower)
                .noreload_h()
                .bit(no_reload_higher)
                .insync()
                .bits(input_sync)
        });
    }
}

/// Set control bits for the entire SCT
pub fn control(
    sct: &pac::SCT,
    mode: CtrlBlock,
    down: bool,
    stop: bool,
    halt: bool,
    bidir: bool,
    pre: Option<u8>,
) -> () {
    unsafe {
        match mode {
            CtrlBlock::Low => {
                sct.ctrl.modify(|_, w| match pre {
                    Some(x) => w
                        .down_l()
                        .bit(down)
                        .stop_l()
                        .bit(stop)
                        .halt_l()
                        .bit(halt)
                        .bidir_l()
                        .bit(bidir)
                        .pre_l()
                        .bits(x)
                        .clrctr_l()
                        .bit(true),
                    _ => w
                        .down_l()
                        .bit(down)
                        .stop_l()
                        .bit(stop)
                        .halt_l()
                        .bit(halt)
                        .bidir_l()
                        .bit(bidir),
                });
            }
            CtrlBlock::High => {
                assert!(!unified!(sct));
                sct.ctrl.modify(|_, w| match pre {
                    Some(x) => w
                        .down_h()
                        .bit(down)
                        .stop_h()
                        .bit(stop)
                        .halt_h()
                        .bit(halt)
                        .bidir_h()
                        .bit(bidir)
                        .pre_h()
                        .bits(x)
                        .clrctr_h()
                        .bit(true),
                    _ => w
                        .down_h()
                        .bit(down)
                        .stop_h()
                        .bit(stop)
                        .halt_h()
                        .bit(halt)
                        .bidir_h()
                        .bit(bidir),
                });
            }
        }
    }
}

/// Configure a match register for the high or low counter
pub fn set_match_reg(sct: &pac::SCT, mode: CtrlBlock, idx: usize, value: u32) {
    assert!(idx < 16, "invalid match register");
    match mode {
        CtrlBlock::Low => unsafe {
            if unified!(sct) {
                sct.match_()[idx].write(|w| w.bits(value));           //Write 32 bits
            } else{
                sct.match_()[idx].modify(|_,w| w.match_l().bits(value as u16));
            }   

        },
        CtrlBlock::High => unsafe {
            assert!(!unified!(sct));
            sct.match_()[idx].modify(|_,w| w.match_h().bits(value as u16));
        },
    }
}

/// Configure a reload register for the high or low counter
pub fn set_reload(sct: &pac::SCT, mode: CtrlBlock, idx: usize, value: u32) {
    assert!(idx < 16);
    unsafe {
        match mode {
            CtrlBlock::Low => {
                if unified!(sct) {
                    sct.matchrel()[idx].write(|w| w.bits(value));           //Write 32 bits
                }else{
                    sct.matchrel()[idx].modify(|_,w| w.reload_l().bits((value & 0xffff) as u16));
                };
            }
            CtrlBlock::High => {
                assert!(!unified!(sct));
                sct.matchrel()[idx].modify(|_,w| w.reload_h().bits((value & 0xffff) as u16));
            }
        }
    }
}

/// read a reload register value for the high or low counter
pub fn get_reload(sct: &pac::SCT, mode: CtrlBlock, idx: usize)-> u32 {
    assert!(idx < 16);
    match mode {
        CtrlBlock::Low => {
            if unified!(sct) {
                return sct.matchrel()[idx].read().bits();                      //Read 32 bits
            }else {
                return sct.matchrel()[idx].read().reload_l().bits() as u32;
            }

        }
        CtrlBlock::High => {
            assert!(!unified!(sct));
            return sct.matchrel()[idx].read().reload_h().bits() as u32;
        }
    }
}

/// Configure the reset limit
pub fn set_limit(sct: &pac::SCT, mode: CtrlBlock, value: u16) {
    match mode {
        CtrlBlock::Low => {
            unsafe {
                sct.limit.modify(|_,w| w.limmsk_l().bits(value));
            }
            //  if unified!(sct) {
            //      unsafe {
            //          sct.limit.write(|w| w.limmsk_h().bits(value as u16));
            //      }
            //  }
        }
        CtrlBlock::High => {
            assert!(!unified!(sct));
            unsafe {
                sct.limit.modify(|_,w| w.limmsk_h().bits(value));
            }
        }
    }
}

/// Set a register in capture control mode or match register mode
///
/// *NOTE:* UM10503 denotes these interchangeably as REGMODE and REGMODEn,
/// here, REGMODE is assumed.
pub fn set_register_mode(sct: &pac::SCT, mode: CtrlBlock, value: u16) {
    match mode {
        CtrlBlock::Low => unsafe {
            sct.regmode.modify(|_,w| w.regmod_l().bits(value));
        },
        CtrlBlock::High => {
            assert!(!unified!(sct));
            unsafe {
                sct.regmode.modify(|_,w| w.regmod_h().bits(value));
            }
        }
    }
}
