pub use embedded_hal::serial::{Read, Write};
// use core::fmt::Write as SerialWrite;
use crate::{
    pac,
    pac::{USART0, USART2, USART3},
};
use nb::Error::Other;
use pac::SCU;

#[allow(unused_imports)]
use crate::io::pin::{
    A, B, C, D, E, F, U0, U1, U10, U11, U12, U13, U14, U15, U16, U17, U18, U19, U2, U20, U3, U4,
    U5, U6, U7, U8, U9,
};

pub struct Serial<USART> {
    pub serial: USART,
}

// Pin typestates
pub struct UartTx<PERIPHERAL>(PhantomData<PERIPHERAL>);
pub struct UartRx<PERIPHERAL>(PhantomData<PERIPHERAL>);

use crate::io::pin::Pin;
use core::marker::PhantomData;

pub trait IntoUartTx<X, N> {
    type Peripheral;
    fn into_uart_tx(self) -> Pin<X, N, UartTx<Self::Peripheral>>;
}

pub trait IntoUartRx<X, N> {
    type Peripheral;
    fn into_uart_rx(self) -> Pin<X, N, UartRx<Self::Peripheral>>;
}

macro_rules! hal {
    ($($USART: ident ($usart: ident): {
        tx: [
            $(($tx_x: tt ($tx_X: ty), $tx_n: tt ($tx_N: ty), $tx_sfs_mode:ident),)+
        ],
        rx: [
            $(($rx_x: tt ($rx_X: ty), $rx_n: tt ($rx_N: ty), $rx_sfs_mode:ident),)+
        ],
    })+) => {
        $(
            $(impl<MODE> IntoUartTx<$tx_X, $tx_N> for Pin<$tx_X, $tx_N, MODE> {
                type Peripheral = $USART;
                // It would have been nice to implement the `From` trait instead, but the blanket
                // `impl<T> From<T> for T` would collide with our use. Since the blanket impl does
                // not use the `default` keyword, it seems we can't even use the gated
                // `specialization` feature.
                fn into_uart_tx(self) -> Pin<$tx_X, $tx_N, UartTx<$USART>> {
                    paste::expr! {
                        unsafe{
                            // Pull-up enabled, Pull-down disabled
                            (*SCU::ptr()).[<sfsp $tx_x _ $tx_n>].write(|w| w
                                    .mode().$tx_sfs_mode()
                                    .epun().clear_bit()
                                    .epd().clear_bit()
                                );
                        }
                    };
                    unsafe { Pin::new() }
                }
            })+

            $(impl<MODE> IntoUartRx<$rx_X, $rx_N> for Pin<$rx_X, $rx_N, MODE> {
                type Peripheral = $USART;
                fn into_uart_rx(self) -> Pin<$rx_X, $rx_N, UartRx<$USART>> {
                    paste::expr! {
                        unsafe{
                            // Pull-up enabled, Pull-down disabled, input buffer enabled
                            (*SCU::ptr()).[<sfsp $rx_x _ $rx_n>].write(|w| w
                                    .mode().$rx_sfs_mode()
                                    .epun().clear_bit()
                                    .epd().clear_bit()
                                    .ezi().set_bit()
                                );
                        }
                    };
                    unsafe { Pin::new() }
                }

            })+

            impl Serial<$USART> {
                pub fn new<Tx, Rx, XTx, NTx, XRx, NRx>(
                    p: $USART,
                    tx_pin: Tx,
                    rx_pin: Rx,
                    dll: u8,
                    dlm: u8,
                    trigger_level: u8
                ) -> Self
                    where Tx: IntoUartTx<XTx, NTx, Peripheral=$USART>,
                          Rx: IntoUartRx<XRx, NRx, Peripheral=$USART>,
                {
                    tx_pin.into_uart_tx();
                    rx_pin.into_uart_rx();
                    assert!(trigger_level < 0x4);
                    let s = Serial {
                        serial: p,
                    };

                    // Setup divisors
                    s.serial.lcr.write(|w| w.dlab().bit(true));
                    unsafe {
                        s.serial.dll().write(|w| w.dllsb().bits(dll));
                        s.serial.dlm().write(|w| w.dlmsb().bits(dlm));
                    }
                    s.serial.lcr.write(|w| w.dlab().bit(false));

                    // Configure FIFOs
                    s.serial.fcr().write(|w| w
                        .fifoen().bit(true)
                        .rxfifores().bit(true)
                        .txfifores().bit(true)
                        .rxtriglvl().bits(trigger_level));

                    s
                }

                /// Reset the RX FIFO
                ///
                /// Clears all elements from the receiver FIFO
                pub fn reset_rx_fifo(&mut self) {
                    self.serial.fcr().write( |w| w.rxfifores().bit(true))
                }

                /// Reset the TX FIFO
                ///
                /// Clears all elements from the transmitter FIFO
                pub fn reset_tx_fifo(&mut self) {
                    self.serial.fcr().write( |w| w.txfifores().bit(true))
                }

                /// Enable/disable interrupts
                pub fn ier(
                    &mut self,
                    rx_data: bool,
                    tx_holding: bool,
                    recv_line: bool,
                    autobaud_end: bool,
                    autobaud_timeout: bool) {
                        self.serial.ier().write(|w| w
                            .rbrie().bit(rx_data)
                            .threie().bit(tx_holding)
                            .rxie().bit(recv_line)
                            .abeointen().bit(autobaud_end)
                            .abtointen().bit(autobaud_timeout));
                    }

                /// Configure frame formatting and break generation
                pub fn lcr(
                    &mut self,
                    wls: pac::$usart::lcr::WLS_A,
                    sbs: pac::$usart::lcr::SBS_A,
                    pe: pac::$usart::lcr::PE_A,
                    ps: pac::$usart::lcr::PS_A,
                    bc: pac::$usart::lcr::BC_A,
                    dlab: pac::$usart::lcr::DLAB_A,
                ) {
                    self.serial.lcr.write(|w| {
                        w.wls()
                            .variant(wls)
                            .sbs()
                            .variant(sbs)
                            .pe()
                            .variant(pe)
                            .ps()
                            .variant(ps)
                            .bc()
                            .variant(bc)
                            .dlab()
                            .variant(dlab)
                    });
                }

                pub fn read_iir(&mut self)->(
                        pac::$usart::iir::INTSTATUS_A , 
                        pac::$usart::iir::INTID_A) {

                    let iir = self.serial.iir().read();
                 
                    let intid = match iir.intid().variant() {
                        pac::generic::Variant::Val(x) => x,
                        _ => panic!("could not get value"),
                    };

                    
                    let intstatus = iir.intstatus().variant();

                    return (intstatus , intid )
                }
            }

            impl core::fmt::Write for Serial<$USART> {
                /// Write a string to the serial port
                ///
                /// Performs a blocking write to output all data in `s` to the serial connection.
                /// If the TX FIFO is non-empty, the write blocks until it is, then it transmits at most 16
                /// bytes before checking the buffer fullness and optionally waiting for it to become empty
                fn write_str(&mut self, s: &str) -> Result<(), core::fmt::Error> {
                    // Wait for output to clear
                    while !self.serial.lsr.read().temt().bit() {}

                    // Dump all characters
                    let mut i: u8 = 0;
                    for c in s.bytes() {
                        // Wait for data to be transmitted every FIFO-length
                        if i >= 15 {
                            // busy wait while data is transmitted
                            while !self.serial.lsr.read().temt().bit() {}
                            i = 0;
                        } else {
                            i += 1;
                        }

                        self.write(c).unwrap();
                    }
                    Ok(())
                }

                /// Write a single byte to the serial port
                ///
                /// Performs a blocking write to output the byte `s`
                fn write_char(&mut self, s: char) -> Result<(), core::fmt::Error> {
                    // Wait for output to clear
                    while !self.serial.lsr.read().temt().bit() {}

                    // Output byte
                    self.write(s as u8).unwrap();
                    Ok(())
                }
            }

            impl Write<u8> for Serial<$USART> {
                type Error = u8;

                fn write(&mut self, data: u8) -> nb::Result<(), Self::Error> {
                    unsafe {
                        self.serial.thr().write(|w| w.thr().bits(data));
                    }

                    let err: u8 = (self.serial.lsr.read().bits() & 0b110001110) as u8;
                    if err != 0 {
                        Err(nb::Error::Other(err))
                    } else {
                        Ok(())
                    }
                }

                fn flush(&mut self) -> nb::Result<(), Self::Error> {
                    while !self.serial.lsr.read().temt().bit() {}
                    Ok(())
                }
            }

            impl Read<u8> for Serial<$USART> {
                type Error = u8;

                fn read(&mut self) -> nb::Result<u8, Self::Error> {
                    let lsr = self.serial.lsr.read().bits() as u8;
                    if lsr != 0 {
                        Ok(self.serial.rbr().read().rbr().bits() as u8)
                    } else {
                        Err(Other(lsr))
                    }
                }
            }
        )+
    }
}

hal! {
    USART0 (usart0): {
        tx: [
            (2 (U2), 0 (U0), function_1),
            (6 (U6), 4 (U4), function_2),
            (9 (U9), 5 (U5), function_7),
            (f (F), 10 (U10), function_1),
        ],
        rx: [
            (2 (U2), 1 (U1), function_1),
            (6 (U6), 5 (U5), function_2),
            (9 (U9), 6 (U6), function_7),
            (f (F), 11 (U11), function_1),
        ],
    }
    USART2 (usart0): {tx: [
        (1 (U1), 15 (U15), function_1),
        (2 (U2), 10 (U10), function_2),
        (7 (U7), 1 (U1), function_6),
        (a (A), 1 (U1), function_3),
    ],
    rx: [
        (1 (U1), 16 (U16), function_1),
        (2 (U2), 11 (U11), function_2),
        (7 (U7), 2 (U2), function_6),
        (a (A), 2 (U2), function_3),
    ],}
    USART3 (usart0): {
        tx: [
            (2 (U2), 3 (U3), function_2),
            (4 (U4), 1 (U1), function_6),
            (9 (U9), 3 (U3), function_7),
            (f (F), 2 (U2), function_1),
        ],
        rx: [
            (2 (U2), 4 (U4), function_2),
            (4 (U4), 2 (U2), function_6),
            (9 (U9), 4 (U4), function_7),
            (f (F), 3 (U3), function_1),
        ],
    }
}
