//! Serial Peripheral Interface abstractions for SPI-capable peripherals.
//!
//! # Which one do I use?
//! The module-level `Spi` uses an SSP peripheral to try to provide an easy-to-use,
//! non-blocking SPI master with full duplex.
//!
//! To implement SPI slaves, the `spi` submodule (using the SPI peripheral) is preferred. This is
//! due to limitations in the interrupts that can be enabled for SSP peripherals.
//!
//! In systems with multiple slaves, it becomes difficult to ensure that transactions finish when
//! using the module-level `Spi` abstraction.  `ssp::SpiSsp` implements blocking transfers, which
//! allows transaction bookkeeping to be done using the existing resource protection of RTFM.

/// Use SSP0 or SSP1 as a SPI master
pub mod ssp;
use ssp::*;

/// Use the SPI peripheral as a SPI slave.
pub mod spi;

/// Manages slave select when using multiple slaves.
pub mod slave_select;

use crate::pac;
pub use embedded_hal::spi::{FullDuplex, Mode, Phase, Polarity};
use heapless::{consts::U17, spsc::Queue};

#[derive(Debug)]
pub enum SpiError {
    ReadKindQueueFull,
    ReadKindQueueEmpty,
    SspError(SspError),
}

fn upgrade_error(e: nb::Error<SspError>) -> nb::Error<SpiError> {
    // not using `From` or `Into` since both types are in the `nb` crate
    match e {
        nb::Error::Other(e) => nb::Error::Other(SpiError::SspError(e)),
        nb::Error::WouldBlock => nb::Error::WouldBlock,
    }
}

fn upgrade_result<T>(r: nb::Result<T, SspError>) -> nb::Result<T, SpiError> {
    // not using `From` or `Into` since both types are in the `nb` crate
    match r {
        Ok(v) => Ok(v),
        Err(e) => Err(upgrade_error(e)),
    }
}

/// Asynchronos abstraction over SPI with a queue to track what to do with the response to each
/// message. The `spi_irq` must be called when the peripheral interrupt occurs.
pub struct Spi<SSP, ReadKind> {
    spi: SpiSsp<SSP>,
    read_queue: heapless::spsc::Queue<ReadKind, U17>,
}

macro_rules! hal {
    ($($SSP: path: ($ssp: ident),)+) => {$(
        impl<ReadKind> Spi<$SSP, ReadKind> {
            /// Configure $SSP for SPI. `initial_read_kind` is the `ReadKind` of the first received
            /// message.
            pub fn new(p: $SSP, mode: Mode, initial_read_kind: ReadKind) -> Self {
                // We don't expose slave configuration for SSP because we don't get interrupts when
                // we receive data.
                let master = true;
                let ssp = SpiSsp::<$SSP>::new(p, mode, master, true);

                // The Tx and Rx FIFOs are length 8. Each message that has been or will be
                // received should have a ReadKind, so we need 16 for the situation when we have a
                // full Rx FIFO and a full Tx FIFO. One additional ReadKind is added for the
                // ReadKind of the first received message, bringing the total to 17.
                let mut queue: Queue<ReadKind, U17> = Queue::new();

                // ReadKind of the first received message.
                queue.enqueue(initial_read_kind).unwrap_or_else(|_err| panic!());

                Spi {
                    spi: ssp,
                    read_queue: queue,
                }
            }

            /// Send a message
            pub fn transact(&mut self, b: u8, read_kind: ReadKind) -> Result<(), nb::Error<SpiError>> {
                match self.read_queue.enqueue(read_kind) {
                    Ok(()) => match self.spi.send(b) {
                        Ok(()) => Ok(()),
                        Err(e) => Err(upgrade_error(e)),
                    },
                    Err(_) => Err(nb::Error::Other(SpiError::ReadKindQueueFull)),
                }
            }

            /// Helper for implementing the interrupt handler. If successful, returns a
            /// `(ReadKind, u8)` pair. The ReadKind should be used to decide what to do with the
            /// read u8.
            pub fn spi_irq(&mut self) -> nb::Result<(ReadKind, u8), SpiError>{
                let b: u8 = upgrade_result(self.spi.read())?;
                let read_kind = self.read_queue.dequeue()
                    .ok_or(nb::Error::Other(SpiError::ReadKindQueueEmpty))?;
                Ok((read_kind, b))
            }

            // TODO expose this in some better way
            pub fn status(&self) -> pac::$ssp::sr::R {
                self.spi.status()
            }

            pub fn receive_not_empty(&self) -> bool {
                self.spi.receive_not_empty()
            }
        }
    )+}
}

hal! {
    pac::SSP0: (ssp0),
    pac::SSP1: (ssp0),
}
