pub use crate::io::pin::SlaveSelect;
pub use paste;

/// Constructs a slave arbiter. Usage:
/// ```
/// spi_arbiter!(ArbiterType (SpiType) {
///     slave_driver_1: Type1,
///     slave_driver_2: Type2,
/// });
///
/// //...
///
/// let spi_arbiter = Arbiter::new(spi, driver_1, driver_2);
///
/// //...
///
/// spi_arbiter.slave_driver_1(|spi, driver_1| {
///     driver_1.do_something_with_slave();
/// });
/// ```
#[macro_export]
macro_rules! spi_arbiter {
    ($Arbiter:ident ($Spi:ident): {$(
        $slave:ident: $ty:ty,
    )+}) => {
        $crate::spi::slave_select::paste::item! {
            /// SPI slave arbiter.
            /// Do not directly access the fields.  Direct access to fields can break the slave
            /// arbitration.  They are visible because of the way the `spi_arbiter` macro inserts
            /// the struct into the user namespace.
            pub struct $Arbiter {
                __internal_spi: $Spi,
                $(
                    [<__internal_ $slave>]: $ty,
                )+
            }

           impl $Arbiter {
               pub fn new(spi: $Spi, $($slave: $ty,)+) -> Self {
                   $Arbiter {
                       __internal_spi: spi,
                       $([<__internal_ $slave>]: $slave,)+
                   }
               }

               $(pub fn $slave<F, T>(&mut self, f: F) -> T
                where F: FnOnce(&mut $Spi, &mut $ty) -> T {
                    self.[<__internal_ $slave>].select();
                    let res = f(&mut self.__internal_spi, &mut self.[<__internal_ $slave>]);
                    self.[<__internal_ $slave>].unselect();
                    res
                })+
           }
        }
    }
}
