//! Implements the non-blocking parts of `embedded_hal` SPI for the SPI peripheral.

use crate::pac;
pub use embedded_hal::blocking::spi::Transfer;
pub use embedded_hal::spi::{FullDuplex, Mode, Phase, Polarity};

#[derive(Debug)]
pub enum Error {
    SlaveAbort,
    ModeFault,
    ReadOverrun,
    WriteCollision,
}

pub struct Spi {
    pub p: pac::SPI,
    // indicates that spif was set, but was cleared it by a read
    transfer_complete: bool,
}

impl Spi {
    /// Configure the SPI peripheral for use.
    /// `master` mode is currently unsupported.
    /// `interrupt` enables interrupts.
    pub fn new(p: pac::SPI, mode: Mode, master: bool, interrupt: bool) -> Self {
        assert!(!master, "Master mode unimplemented. Use the SSP.");
        p.ccr.write(|w| unsafe {
            // spi_sck = pclk_spi / counter
            //
            // requires
            // counter >= 8 and counter is even
            w.counter().bits(8)
        });

        // this clears Slave abort and Read overrun status and sets us up to clear the rest
        p.sr.read();

        // clears Write collision and transfer complete status
        p.dr.read();

        // clears Mode fault status
        p.cr.write(|w| {
            w.bitenable()
                .clear_bit() // set to use the `bits`
                .cpha()
                .bit(mode.phase == Phase::CaptureOnSecondTransition)
                .cpol()
                .bit(mode.polarity == Polarity::IdleHigh)
                .mstr()
                .bit(master)
                .lsbf()
                .msb() // MSB or LSB transferred first?
                .spie()
                .bit(interrupt) // enable interrupt
        });

        Spi {
            p,
            transfer_complete: true,
        }
    }
}

impl FullDuplex<u8> for Spi {
    type Error = Error;

    fn read(&mut self) -> nb::Result<u8, Error> {
        let sr = self.p.sr.read();

        if sr.abrt().bit_is_set() {
            Err(nb::Error::Other(Error::SlaveAbort))
        } else if sr.modf().bit_is_set() {
            Err(nb::Error::Other(Error::ModeFault))
        } else if sr.rovr().bit_is_set() {
            Err(nb::Error::Other(Error::ReadOverrun))
        } else if sr.wcol().bit_is_set() {
            Err(nb::Error::Other(Error::WriteCollision))
        } else if sr.spif().bit_is_set() {
            self.transfer_complete = true;
            Ok(self.p.dr.read().datalow().bits())
        } else {
            Err(nb::Error::WouldBlock)
        }
    }

    fn send(&mut self, b: u8) -> nb::Result<(), Error> {
        let sr = self.p.sr.read();

        if sr.abrt().bit_is_set() {
            Err(nb::Error::Other(Error::SlaveAbort))
        } else if sr.modf().bit_is_set() {
            Err(nb::Error::Other(Error::ModeFault))
        } else if sr.rovr().bit_is_set() {
            Err(nb::Error::Other(Error::ReadOverrun))
        } else if sr.wcol().bit_is_set() {
            Err(nb::Error::Other(Error::WriteCollision))
        } else if sr.spif().bit_is_set() || self.transfer_complete {
            // spif can be cleared by reading or writing the data register, so we have to do some
            // bookkeeping of our own
            self.transfer_complete = false;
            self.p.dr.write(|w| unsafe { w.datalow().bits(b) });
            Ok(())
        } else {
            Err(nb::Error::WouldBlock)
        }
    }
}

impl embedded_hal::blocking::spi::transfer::Default<u8> for Spi {}

// Will get read overrun ROVR on second write unless it is ignored.
// need a custom impl that ignores ROVR if we want to use `Write`
//impl embedded_hal::blocking::spi::write::Default<u8> for Spi {}
