use crate::pac;
pub use embedded_hal::spi::{FullDuplex, Mode, Phase, Polarity};

/// Use an SSP peripheral for SPI. Keeping track of what data is being read is left to the user.
/// Take care when using the `read` method of `FullDuplex`. This will read the head of the receive
/// FIFO, and it is up to the user to keep track of which transmission the data comes from.
///
/// The blocking `embedded_hal::blocking::spi::Transfer` can be used if blocking behavior is
/// acceptable.
pub struct SpiSsp<SSP> {
    p: SSP,
}

#[derive(Debug)]
pub enum SspError {}

macro_rules! hal_ssp {
    ($($SSP: path: ($ssp: ident, $ssp_rst: ident),)+) => {
        $(
            impl SpiSsp<$SSP> {
                /// Configure the SSP peripheral for SPI.
                /// Slave mode is currently not well-supported.
                pub fn new(p: $SSP, mode: Mode, master: bool, interrupt: bool) -> Self {
                    p.cr0.write(|w| {
                        w.dss()
                            ._8_bit_transfer()
                            .frf()
                            .spi()
                            .cpol()
                            .bit(mode.polarity == Polarity::IdleHigh) // bus clock state between frames
                            .cpha()
                            .bit(mode.phase == Phase::CaptureOnSecondTransition)
                    });


                    // clock prescale register
                    p.cpsr.write(|w| unsafe { w.cpsdvsr().bits(16) });

                    if interrupt {
                        // interrupt mask
                        p.imsc.write(|w| {
                            w.rorim()
                                .clear_bit() // Receive Overrun: Rx FIFO is full and another frame is completely received
                                .rtim()
                                .set_bit() // Receive Time-out
                                .rxim()
                                .set_bit() // Rx FIFO is at least half full
                                .txim()
                                .clear_bit() // Tx FIFO is at least half empty
                        });
                    } else {
                        // reset interrupt mask
                        p.imsc.write(|w| w);
                    }


                    // enable the SSP controller last
                    p.cr1.write(|w| {
                        w
                            .sse()
                            .enabled() // enable the SSP controller
                            .ms()
                            .bit(!master) // slave mode when 1
                    });

                    if !master {
                        // prepare an initial (garbage) response for slave
                        p.dr.write(|w| unsafe { w.data().bits(42u16) });
                    }

                    SpiSsp { p }
                }

                // TODO expose status in some better way
                pub fn status(&self) -> pac::ssp0::sr::R {
                    self.p.sr.read()
                }

                /// True if there is data to read.
                pub fn receive_not_empty(&self) -> bool {
                    self.p.sr.read().rne().bit_is_set()
                }
            }

            impl FullDuplex<u8> for SpiSsp<$SSP> {
                type Error = SspError;

                fn read(&mut self) -> nb::Result<u8, SspError> {
                    if self.p.sr.read().rne().bit_is_set() {
                        let b = self.p.dr.read().data().bits() as u8; // u16
                        Ok(b)
                    } else {
                        Err(nb::Error::WouldBlock)
                    }
                }

                fn send(&mut self, b: u8) -> nb::Result<(), SspError> {
                    // using 'Transmit FIFO Empty' instead of 'Transmit FIFO Not Full' to support
                    // per-message size selection
                    if self.p.sr.read().tfe().bit_is_set() {
                        self.p.cr0.modify(|_, w| w.dss()._8_bit_transfer());
                        self.p.dr.write(|w| unsafe { w.data().bits(b as u16) });
                        Ok(())
                    } else {
                        Err(nb::Error::WouldBlock)
                    }
                }
            }

            impl FullDuplex<u16> for SpiSsp<$SSP> {
                type Error = SspError;

                fn read(&mut self) -> nb::Result<u16, SspError> {
                    if self.p.sr.read().rne().bit_is_set() {
                        let b = self.p.dr.read().data().bits();
                        Ok(b)
                    } else {
                        Err(nb::Error::WouldBlock)
                    }
                }

                fn send(&mut self, b: u16) -> nb::Result<(), SspError> {
                    // using 'Transmit FIFO Empty' instead of 'Transmit FIFO Not Full' to support
                    // per-message size selection
                    if self.p.sr.read().tfe().bit_is_set() {
                        self.p.cr0.modify(|_, w| w.dss()._16_bit_transfer());
                        self.p.dr.write(|w| unsafe { w.data().bits(b) });
                        Ok(())
                    } else {
                        Err(nb::Error::WouldBlock)
                    }
                }
            }

            impl embedded_hal::blocking::spi::transfer::Default<u8> for SpiSsp<$SSP> {}
            impl embedded_hal::blocking::spi::transfer::Default<u16> for SpiSsp<$SSP> {}

            // This will fill up and overrun the read FIFO, but it could be useful for use cases
            // where the master _never_ wants to read.
            //impl embedded_hal::blocking::spi::write::Default<u8> for SpiSsp<$SSP> {}
        )+
    }
}

hal_ssp! {
    pac::SSP0: (ssp0, ssp0_rst),
    pac::SSP1: (ssp0, ssp1_rst),
}
