use crate::pac;
use pac::{
    SCU, TIMER0, TIMER1, TIMER2, TIMER3,
};
pub use pac::timer0::{
    ctcr::{CINSEL_A, CTMODE_A},
    emr::{EMC0_A, EMC1_A, EMC2_A, EMC3_A},
    mcr::*,
};

#[allow(unused)]
use crate::io::pin::{ U0, U1, U2, U3, U4, U5, U6, U7, U8, U9, U10, U11, U12, U13, U14, U15, U16, U17, U18, U19, U20, C, };
use crate::io::pin::{Pin, Uninit};
use core::marker::PhantomData;

pub struct Timer<TIMER, O, P, Q, R, S, T, U, V>
    where
        O: IntoMatch<TIMER, Match0>,
        P: IntoMatch<TIMER, Match1>,
        Q: IntoMatch<TIMER, Match2>,
        R: IntoMatch<TIMER, Match3>,
        S: IntoCapture<TIMER, Capture0>,
        T: IntoCapture<TIMER, Capture1>,
        U: IntoCapture<TIMER, Capture2>,
        V: IntoCapture<TIMER, Capture3>,
{
    timer: TIMER,
    _match0_pin: PhantomData<O>,
    _match1_pin: PhantomData<P>,
    _match2_pin: PhantomData<Q>,
    _match3_pin: PhantomData<R>,
    _capture0: PhantomData<S>,
    _capture1: PhantomData<T>,
    _capture2: PhantomData<U>,
    _capture3: PhantomData<V>,
}

/// Match register 0
pub struct Match0;
/// Match register 1
pub struct Match1;
/// Match register 2
pub struct Match2;
/// Match register 3
pub struct Match3;

/// Capture register 0
pub struct Capture0;
/// Capture register 1
pub struct Capture1;
/// Capture register 2
pub struct Capture2;
/// Capture register 3
pub struct Capture3;

/// Match register output
pub struct Match<PERIPHERAL, CHANNEL> {
    _peripheral: PhantomData<PERIPHERAL>,
    _channel: CHANNEL,
}

/// Timer capture input
pub struct Capture<PERIPHERAL, CHANNEL> {
    _peripheral: PhantomData<PERIPHERAL>,
    _channel: CHANNEL,
}

/// `TIMER` is the HAL timer type. `C` is the channel that the match pin will respond to
pub trait IntoMatch<TIMER, C> {
    /// Output pin type
    type P;

    /// Convert pin to HAL type
    fn into_match(self) -> Self::P;
}


pub struct NoMatchPin {}
impl NoMatchPin {
    pub fn new() -> Self {
        NoMatchPin{}
    }
}
impl<TIMER, C> IntoMatch<TIMER, C> for NoMatchPin {
    type P = Self;
    fn into_match(self) -> Self {
        self
    }
}

pub struct NoCapturePin {}
impl NoCapturePin {
    pub fn new() -> Self {
        NoCapturePin{}
    }
}
impl<TIMER, C> IntoCapture<TIMER, C> for NoCapturePin {
    type P = Self;
    unsafe fn into_capture(self) -> Self {
        self
    }
}

pub trait IntoCapture<TIMER, C> {
    /// Output pin type
    type P;
    /// Reconfigure the pin for capture on timer `TIMER` and channel `C`.
    ///
    /// NOTE: Implementers are currently not required to configure GIMA, which will be necessary
    /// for the peripheral to get the capture. Instead, the user must configure GIMA manually.
    unsafe fn into_capture(self) -> Self::P;
}

#[derive(PartialEq)]
pub enum InterruptType {
    MatchChannel,
    CaptureChannel,
}

macro_rules! hal {
    ($($TIMER: ident: ($timer: ident),)+) => {
        $(
            impl<A, B, C, D, E, F, G, H> Timer<$TIMER, A, B, C, D, E, F, G, H>
                where
                    A: IntoMatch<$TIMER, Match0>,
                    B: IntoMatch<$TIMER, Match1>,
                    C: IntoMatch<$TIMER, Match2>,
                    D: IntoMatch<$TIMER, Match3>,
                    E: IntoCapture<$TIMER, Capture0>,
                    F: IntoCapture<$TIMER, Capture1>,
                    G: IntoCapture<$TIMER, Capture2>,
                    H: IntoCapture<$TIMER, Capture3>,
            {
                /// Create a new instance of the interface
                pub fn new(device: $TIMER,
                           match0_pin: Option<A>,
                           match1_pin: Option<B>,
                           match2_pin: Option<C>,
                           match3_pin: Option<D>,
                           capture0: Option<E>,
                           capture1: Option<F>,
                           capture2: Option<G>,
                           capture3: Option<H>,
                ) -> Timer<$TIMER, A, B, C, D, E, F, G, H> {
                    match0_pin.map(|p| p.into_match());
                    match1_pin.map(|p| p.into_match());
                    match2_pin.map(|p| p.into_match());
                    match3_pin.map(|p| p.into_match());
                    unsafe {
                        capture0.map(|p| p.into_capture());
                        capture1.map(|p| p.into_capture());
                        capture2.map(|p| p.into_capture());
                        capture3.map(|p| p.into_capture());
                    }

                    Timer {
                        timer: device,
                        _match0_pin: PhantomData,
                        _match1_pin: PhantomData,
                        _match2_pin: PhantomData,
                        _match3_pin: PhantomData,
                        _capture0: PhantomData,
                        _capture1: PhantomData,
                        _capture2: PhantomData,
                        _capture3: PhantomData,
                    }
                }

                /// Create a new instance of the interface, with no pins configured.
                pub fn new_no_pins(device: $TIMER) -> Self {
                    Self::new(device, None, None, None, None, None, None, None, None)
                }

                /// Start the timer
                ///
                /// Return the timer from an idle state to a running state, if
                /// it was being reset it is brought out of it
                pub fn start_timer (&mut self, runstop: bool) {
                    self.timer.tcr.modify(|_, w| w.cen().bit(runstop));

                    if runstop {
                        self.timer.tcr.modify(|_, w| w.crst().bit(false));
                    }
                }

                /// Reset the timer
                ///
                /// Pulls the timer reset flag high and keeps it there until
                /// `start_timer` brings it back down
                pub fn reset_timer (&mut self) {
                    self.timer.tcr.write(|w| w.crst().bit(true));
                }

                /// Sets a match register for channel `$idx` in `$timer` to `$val`
                ///
                /// The match registers are continuously compared to the timer counter and
                /// trigger events whenever they match. For instance: A 50% duty cycle PWM
                /// signal could be generated by using `timer_set_mcr!` to reset the timer
                /// counter at `PWM_PERIOD` and at the same time set an output pin low while
                /// a second match register would be set at `PWM_PERIOD/2` and assert that very
                /// same pin high.
                ///
                /// # Examples
                /// For a related example, check the documentation on `timer_set_external!`
                pub fn set_match (&mut self, idx: usize, val: u32) {
                    unsafe {
                        assert!(idx < 4);
                        self.timer.mr[idx].write(|w| w.match_().bits(val));
                    }
                }

                /// Set the prescaler and prescaler counter
                pub fn set_prescaler (&mut self, pr: u32, pc: Option<u32>) {
                    unsafe {
                        self.timer.pr.write(|w| w.pm().bits(pr));
                        self.timer.pc.write(|w| w.pc().bits(match pc {
                            Some(x) => x,
                            None => 0
                        }))
                    }
                }

                /// Clear the timer interrupt
                pub fn clear_interrupt (&mut self, int: InterruptType, idx: u8) {
                    unsafe {
                        if int == InterruptType::MatchChannel {
                            self.timer.ir.write(|w| w.bits(1<<idx));
                        } else {
                            self.timer.ir.write(|w| w.bits(1<<(idx+3)));
                        }
                    }
                }

                pub fn set_ctmode (&mut self, ctmode: CTMODE_A, cinsel: CINSEL_A) {
                    self.timer.ctcr.write(|w| w
                        .ctmode().variant(ctmode)
                        .cinsel().variant(cinsel));
                }

                /// Reads the timer count register
                pub fn read_tc(&self) -> u32 {
                    return self.timer.tc.read().bits();

                }

                mr_impls!(0, 1, 2, 3,);
            }
        )+
    }
}

macro_rules! mr_impls {
    ($($x: tt,)+) => {
        paste::item! {
            $(
                /// Configure behavior for match register $x
                ///
                /// Safety: If the timer is running, the EMR read-modify-write will cause a race.
                pub unsafe fn [<mr $x>](&mut self, interrupt: [<MR $x I_A>], reset: [<MR $x R_A>], stop: [<MR $x S_A>], match_action: [<EMC $x _A>]) {
                    self.timer.mcr.modify(|_, w| w
                        .[<mr $x i>]().variant(interrupt)
                        .[<mr $x r>]().variant(reset)
                        .[<mr $x s>]().variant(stop));
                    self.timer.emr.modify(|_, w| w.[<emc $x>]().variant(match_action));
                }

                /// Set External Match $x. When a match occurs between the TC and [<MR $i>], this bit can
                /// either toggle, go low, go high, or do nothing, depending on the mode set using
                /// `[<mr $x>](.., emc)`. This is used to overwrite the output when a match output pin is
                /// used.
                ///
                /// ## Safety
                /// The read-modify-write of `emr` can happen concurrently with the peripheral
                /// modifying the external match due to a match event.
                /// This is especially likely to happen when the duty cycle is very close to 0%.
                /// This will produce close to 100% duty cycle for a single PWM period each time
                /// the case comes up.
                ///
                /// If the timer is stopped, this is trivially safe.
                ///
                /// Otherwise some other argument needs to be made about the timings of timer
                /// matches and `emr` accesses. This should typically involve a comparison of
                /// the maximum time between a timer interrupt being asserted in the NVIC and the
                /// running this function and the timing of a timer match.
                pub unsafe fn [<write_em $x>](&mut self, high: bool) {
                    self.timer.emr.modify(|_, w| w.[<em $x>]().bit(high));
                }

            )+
        }
    }
}

hal! {
    TIMER0: (timer0),
    TIMER1: (timer1),
    TIMER2: (timer2),
    TIMER3: (timer3),
}

macro_rules! match_pins {
    ($($TIMER: ident: {
        $($x:tt ($X:ty), $n:tt ($N:ty), $sfs_mode: ident, $c: tt, $C: ty;)+
    },)+) => {
        paste::item! {
            $(
                $(
                    impl<TIMER> IntoMatch<TIMER, $C> for Pin<$X, $N, Uninit> {
                        type P = Pin<$X, $N, Match<TIMER, C>>;

                        fn into_match(self) -> Self::P {
                            // Configure pin state
                            // This is safe because we were given the pin to work with
                            unsafe {
                                (*SCU::ptr()).[<sfsp $x _ $n>].write(|w| w.mode().$sfs_mode());
                            }

                            // Safe because we got the pin as an input
                            unsafe {
                                Self::P::new()
                            }
                        }
                    }
                )+
            )+
        }
    }
}

macro_rules! capture_pins {
    ($($TIMER: ident: {
        $($x:tt ($X:ty), $n:tt ($N:ty), $sfs_mode: ident, $c: tt, $C: ty;)+
    },)+) => {
        paste::item! {
            $(
                $(
                    impl<TIMER> IntoCapture<TIMER, $C> for Pin<$X, $N, Uninit> {
                        type P = Pin<$X, $N, Capture<TIMER, C>>;

                        unsafe fn into_capture(self) -> Self::P {
                            // Configure pin state
                            //
                            // Safety
                            // Because we own `self`, a `Pin<X, N>`, we are guaranteed exclusive access to `sfspx_n`
                            (*SCU::ptr()).[<sfsp $x _ $n>].write(|w| w
                                .mode().$sfs_mode()
                                .ezi().enable_input_buffer()
                            );

                            // Safe because the output `(X, N)` are the same as the input `(X, N)`
                            Self::P::new()
                        }
                    }
                )+
            )+
        }
    }
}

match_pins! {
   TIMER0: {
       1 (U1), 16 (U16), function_4, 0, Match0;
       1 (U1), 15 (U15), function_4, 1, Match1;
       1 (U1), 14 (U14), function_4, 2, Match2;
       1 (U1), 18 (U18), function_4, 3, Match3;

       8 (U8), 0 (U0), function_7, 0, Match0;
       8 (U8), 1 (U1), function_7, 1, Match1;
       8 (U8), 2 (U2), function_7, 2, Match2;
       8 (U8), 3 (U3), function_7, 3, Match3;
   },
    TIMER1: {
        5 (U5), 4 (U4), function_5, 0, Match0;
        5 (U5), 5 (U5), function_5, 1, Match1;
        5 (U5), 6 (U6), function_5, 2, Match2;
        5 (U5), 7 (U7), function_5, 3, Match3;
    },
   TIMER2: {
       6 (U6), 7 (U7),   function_5, 0, Match0;
       6 (U6), 8 (U8),   function_5, 1, Match1;
       6 (U6), 9 (U9),   function_5, 2, Match2;
       6 (U6), 11 (U11), function_5, 3, Match3;
   },
   TIMER3: {
       2 (U2), 3 (U3), function_6, 0, Match0;
       2 (U2), 4 (U4), function_6, 1, Match1;
       2 (U2), 5 (U5), function_6, 2, Match2;
       2 (U2), 7 (U7), function_6, 3, Match3;

       c (C), 7  (U7),  function_6, 0, Match0;
       c (C), 8  (U8),  function_6, 1, Match1;
       c (C), 9  (U9),  function_6, 2, Match2;
       c (C), 10 (U10), function_6, 3, Match3;
   },
}

// Right now, we only encode the choices for Tx_CAPn, not the other GIMA possibilities. This should be
// extended.
capture_pins! {
    TIMER0: {
        1 (U1), 13 (U13), function_4, 0, Capture0;
        1 (U1), 12 (U12), function_4, 1, Capture1;
        1 (U1), 20 (U20), function_4, 2, Capture2;
        1 (U1), 17 (U17), function_4, 3, Capture3;
        8 (U8), 4 (U4), function_7, 0, Capture0;
        8 (U8), 5 (U5), function_7, 1, Capture1;
        8 (U8), 6 (U6), function_7, 2, Capture2;
        8 (U8), 7 (U7), function_7, 3, Capture3;
    },
    TIMER1: {
        5 (U5), 0 (U0), function_5, 0, Capture0;
        5 (U5), 1 (U1), function_5, 1, Capture1;
        5 (U5), 2 (U2), function_5, 2, Capture2;
        5 (U5), 3 (U3), function_5, 3, Capture3;
    },
    TIMER2: {
        6 (U6), 1 (U1), function_5, 0, Capture0;
        6 (U6), 2 (U2), function_5, 1, Capture1;
        6 (U6), 3 (U3), function_5, 2, Capture2;
        6 (U6), 6 (U6), function_5, 3, Capture3;
    },
    TIMER3: {
        2 (U2), 0 (U0), function_6, 0, Capture0;
        2 (U2), 1 (U1), function_6, 1, Capture1;
        2 (U2), 2 (U2), function_6, 2, Capture2;
        2 (U2), 6 (U6), function_6, 3, Capture3;
        c (C), 1 (U1), function_6, 0, Capture0;
        c (C), 4 (U4), function_6, 1, Capture1;
        c (C), 5 (U5), function_6, 2, Capture2;
        c (C), 6 (U6), function_6, 3, Capture3;
    },
}
